import numpy as np
_float = float
from semirings.base import Semiring, Chart, Wrapped, make_semiring
from semirings.mert import ConvexHull, Point
from semirings.lazysort import LazySort, flatten
from semirings.maxplus import MaxPlus
from semirings.maxtimes import MaxTimes
from semirings.mintimes import MinTimes
from semirings.minplus import MinPlus
from semirings.logval import LogVal
from semirings.float import Float
from semirings.boolean import Boolean
from semirings.lazysort import LazySort
from semirings.cutsets import CutSets
from semirings.regex import Symbol
from semirings.interval import Interval, make_interval
from semirings.misc import *


Dual = dual(Float)



"""
MERT semiring
"""
import numpy as np
import pylab as pl
import scipy.spatial
from semirings import Semiring

# TODO: there is nothing specific about the convexhull semiring and being
# two-dimensional.

# TODO: I read somewhere that the union of convex hulls can be computed in
# linear time (no log factor).

# TODO: convex hull operation commutes with minkowski sum (no need for a second
# call to convex hull)

class ConvexHull(Semiring):
    """Convex hull semiring (in two-dimensions).

    Each element of this semiring is a convex hull of `Points`.

    Based closely on Chris Dyer's 2013 arxiv paper
    (http://arxiv.org/pdf/1307.3675.pdf)

    Coded for clarity, not efficiency.

    """

    def __init__(self, points):
        self.points = conv(points)

    def __iter__(self):
        return iter(self.points)

    def __eq__(self, other):
        return set(self.points) == set(other.points)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.points})'

    def __hash__(self):
        return hash(frozenset(self.points))

    def __add__(self, other):
        assert isinstance(other, ConvexHull)
        return ConvexHull(list(self) + list(other))

    def __mul__(self, other):
        # http://en.wikipedia.org/wiki/Minkowski_addition
        assert isinstance(other, ConvexHull)
        return ConvexHull([a + b
                           for a in self
                           for b in other])

    @classmethod
    def multiplicity(cls, v, m):
        return v if m > 0 else cls.zero

    # The star operation diverges
    def star(self):
        #return self.one + self
        raise NotImplementedError()

    def draw(self, label_fmt=None, c='k'):   # pragma: no cover
        "Visualize points with interactive scatter plot browser."
        if len(self.points) <= 2:
            print('[warn] ConvexHull has too few points empty.')
            return

        points = np.array([(p.x, p.y) for p in self.points])
        hull = scipy.spatial.ConvexHull(points, qhull_options='QJ Pp')
        for simplex in hull.simplices:
            pl.plot(points[simplex, 0], points[simplex, 1], zorder=-1, alpha=0.5, lw=.5, color=c)

        for p in self:
            pl.scatter(p.x, p.y, alpha=0.5, zorder=-1, c=c)

        pl.box(False)
        pl.axis('equal')
        pl.xticks([p.x for p in self], rotation='vertical')
        pl.yticks([p.y for p in self])
        for p in self:
            pl.text(x=p.x, y=p.y, s=str(p.d) if label_fmt is None else label_fmt(p))


def conv(points, jiggle=True):
    "Indices of the convex hull."
    if len(points) <= 2:
        return points
    else:
        c = scipy.spatial.ConvexHull(
            np.array([(x.x, x.y) for x in points]),
            qhull_options=('QJ Pp' if jiggle else None)
        )
        return [points[i] for i in c.vertices]


class Point:
    """
    Two-dimensional point with backpointers so that we can reconstruct the
    derivation.
    """

    def __init__(self, x, y, d=None):
        self.x = x
        self.y = y
        self.d = d

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __hash__(self):
        return hash((self.x, self.y))

    def __add__(self, other):
        return Point(self.x + other.x,
                     self.y + other.y,
                     (self, other))

    def __repr__(self):
        #d = '' if self.d is None else f', {self.d}'
        #d = t._pformat_flat(nodesep='', parens='()', quotes=False)
        #return f'Point({self.x}, {self.y}{d})'
        return str(self.d)


zero = ConvexHull([])
one = ConvexHull([Point(0,0)])

ConvexHull.zero = zero
ConvexHull.one = one



from . import base


class Boolean(base.Semiring):

    def __init__(self, score, d=None):
        self.score = (score > 0)
        self.d = d

    def approx_zero(self):
        return not self

    def __add__(self, other):
        return max(self, other)

    def __mul__(self, other):
        if other is one: return self
        if self is one: return other
        if other is zero: return zero
        if self is zero: return zero
        return Boolean(self.score and other.score, [self.d, other.d])

    def __hash__(self):
        return hash(self.score)

    def __eq__(self, other):
        return isinstance(other, Boolean) and self.score == other.score

    def __lt__(self, other):
        return self.score < other.score

    def __repr__(self):
#        return f'Boolean({self.score}, {self.d})'
        return f'{self.score}'

    def star(self):
        return one

    @classmethod
    def lift(cls, score, d):
        #if isinstance(score, Boolean):
        #    return score
        #else:
        return cls(score, d)

    @classmethod
    def multiplicity(cls,v,m):
        return v if m > 0 else cls.zero

    def __bool__(self):
        return self.score


Boolean.zero = zero = Boolean(False, None)
Boolean.one = one = Boolean(True, ())



import numpy as np
from arsenal.iterextras import sorted_union, sorted_product, take
from . import base

# TODO: create a LazySort builder which is parametric in the underlying ops,
# just like expectation semiring.  This will reuse code for max-sum, min-sum,
# sum-prod, and even lexicographic semirings.

class _LazySort(base.Semiring):
    def __add__(self, other):
        if self is zero: return other
        if other is zero: return self
        return Sum(self, other)
    def __mul__(self, other):
        if self is one: return other
        if other is one: return self
        if self is zero: return zero
        if other is zero: return zero
        return Prod(self, other)
    def star(self):
        if self is zero: return one
        return Star(self)
    @classmethod
    def multiplicity(cls, v, m):
        return cls.zero if m == 0 else Multiplicity(v, m)
    def take(self, K):
        return take(K, self)


class Multiplicity(_LazySort):
    def __init__(self, a, m):
        assert isinstance(a, _LazySort), a
        self.a = a
        self.m = m
        super().__init__()
    def __iter__(self):
        for v in self.a:
            i = 0
            while i < self.m:
                yield v
                i += 1
    def __repr__(self):
        return f'multiplicity({self.a},{self.m})'


# Notice that in this multiplication is neither associative, nor commutative
# because of the backpointers. This is how we get out a derivation tree (with
# parentheses).  If we ignore backpointers, then we are are associative.
class LazySort(_LazySort):
    def __init__(self, score, data):
        #assert isinstance(score, (int,float)), score
        self.score = score
        self.data = data
        super().__init__()

    # TODO: it is a strange choice to have a different product operation than
    # _LazySort - can we change that?  I think the issue is actually that there
    # is a "base" product that is passed to sorted_product and all that.  And,
    # we've sort of conflated things when we should have kept them separate.
    # The objects that we iterate over should be instances of an ordered base
    # semiring.  The details of backpointers (like we see below) would be
    # handled there too.
    def __mul__(self, other):
        if self is one: return other
        if other is one: return self
        if self is zero: return zero
        if other is zero: return zero
        if isinstance(other, LazySort):
            return LazySort(self.score * other.score, [self.data, other.data])
        else:
            return super().__mul__(other)
    def __lt__(self, other):
        return other.score < self.score    # Warning: this is backwards!
    def __iter__(self):
        yield self
    def __eq__(self, other):
        return isinstance(other, LazySort) and self.score == other.score
    def __repr__(self):
        return repr((self.score, self.data))
    def flat_data_list(self):
        return flatten(self.data)


class Sum(_LazySort):
    def __init__(self, a, b):
        assert isinstance(a, _LazySort) and isinstance(b, _LazySort), [a,b]
        self.a = a
        self.b = b
        super().__init__()
    def __iter__(self):
        yield from sorted_union(self.a, self.b)
    def __repr__(self):
#        return f'({self.a} + {self.b})'
        return 'sum(...)'


class Prod(_LazySort):
    def __init__(self, a, b):
        assert isinstance(a, _LazySort) and isinstance(b, _LazySort), [a,b]
        self.a = a
        self.b = b
        super().__init__()
    def __iter__(self):
        yield from sorted_product(np.prod, self.a, self.b)
    def __repr__(self):
        return f'({self.a} * {self.b})'


class Star(_LazySort):
    def __init__(self, a):
        assert isinstance(a, _LazySort), a
        self.a = a
        super().__init__()
    def __iter__(self):
        if self.a is zero:
            yield one
            return
        v = one
        while True:
            print('star...', self.a)
            yield from v
            v *= self.a
    def __repr__(self):
        return f'star({self.a})'


class Zero(_LazySort):
    def __init__(self):
        self.score = 0
        super().__init__()
    def __iter__(self):
        if 0: yield
    def __repr__(self):
        return 'NULL'


zero = Zero()
one = LazySort(1, ())

LazySort.zero = zero
LazySort.one = one


def flatten(S):
    if not isinstance(S, list):
        return [S]
    else:
        tmp = []
        for x in S:
            tmp.extend(flatten(x))
        return tmp



# TODO: fsa should move into utilities since it is not a semiring - it is a data
# structure that supports the semiring for formal languages
from arsenal import Integerizer, colors
from collections import defaultdict
from functools import lru_cache


def dfs(Ps, arcs):
    stack = list(Ps)
    m = FSA()
    for P in Ps: m.add_start(P)
    while stack:
        P = stack.pop()
        for a, Q in arcs(P):
            if Q not in m.nodes:
                stack.append(Q)
                m.nodes.add(P)
            m.add(P, a, Q)
    return m


_frozenset = frozenset
class frozenset(_frozenset):
    def __repr__(self):
        return '{%s}' % (','.join(str(x) for x in sorted(self)))


class FSA:

    def __init__(self):
        self.start = set()
        self.edges = defaultdict(lambda: defaultdict(set))
        self.nodes = set()
        self.stop = set()
        self.syms = set()

    def as_tuple(self):
        return (frozenset(self.nodes),
                frozenset(self.start),
                frozenset(self.stop),
#                frozenset(self.syms),
                frozenset(self.arcs()))

    def __hash__(self):
        return hash(self.as_tuple())

    def __eq__(self, other):
        return self.as_tuple() == other.as_tuple()

    def __repr__(self):
#        return f'<{self.__class__.__name__} id={id(self)}>'
        return repr(self.to_regex())

    def __str__(self):
        x = ['{']   # todo: better print; include start/stop
        for s in sorted(self.nodes):
            ss = f'{s}'
            if s in self.start:
                ss = f'^{ss}'
            if s in self.stop:
                ss = f'{ss}$'
            x.append(f'  {ss}:')
            for a, t in sorted(self.arcs(s)):
                x.append(f'    {a} -> {t}')
        x.append('}')
        return '\n'.join(x)

    def _repr_html_(self):
        return self.graphviz()._repr_image_svg_xml()

    def graphviz(self, show_label=True):
        from graphviz import Digraph
        g = Digraph(
            graph_attr=dict(rankdir='LR'),
            node_attr=dict(
                fontname='Monospace',
                fontsize='10',
                height='.05', width='.05',
                #margin="0.055,0.042"
                margin="0,0"
            ),
            edge_attr=dict(
                #arrowhead='vee',
                arrowsize='0.3',
                fontname='Monospace',
                fontsize='9'
            ),
        )
        f = Integerizer()

        # FIXME: make sure this name is actually unqiue
        start = '<start>'
        g.node(start, label='', shape='point', height='0', width='0')
        for i in self.start:
            g.edge(start, str(f(i)), label='')

        for i in sorted(self.nodes):
            shape = 'circle'
            if i in self.stop: shape = 'doublecircle'
            label = str(i) if show_label else ''
            #if i in self.start: label = '*'
            g.node(str(f(i)), label=label, shape=shape)
            for a, j in sorted(self.arcs(i)):
                g.edge(str(f(i)), str(f(j)), label=str(a))

        return g

    def D(self, x):
        "left derivative"
        m = FSA()

        e = self.epsremoval()
        for i,a,j in e.arcs():

            if i in e.start and a == x:
                m.add(i,eps,j)
            else:
                m.add(i,a,j)

        m.start = set(e.start)
        m.stop = set(e.stop)
        return m

    def add(self, i, a, j):
        self.edges[i][a].add(j)
        self.nodes.add(i); self.syms.add(a); self.nodes.add(j)
        return self

    def add_start(self, i):
        self.start.add(i)
        self.nodes.add(i)
        return self

    def add_stop(self, i):
        self.stop.add(i)
        self.nodes.add(i)
        return self

    def arcs(self, i=None, a=None):
        if i is None and a is None:

            for i in self.edges:
                for a in self.edges[i]:
                    for j in self.edges[i][a]:
                        yield (i,a,j)

        elif i is not None and a is None:

            for a in self.edges[i]:
                for j in self.edges[i][a]:
                    yield (a,j)

        elif i is not None and a is not None:

            for j in self.edges[i][a]:
                yield j

        else:
            raise NotImplementedError()

    def reverse(self):
        m = FSA()
        for i in self.start:
            m.add_stop(i)
        for i in self.stop:
            m.add_start(i)
        for i, a, j in self.arcs():
            m.add(j, a, i)     # pylint: disable=W1114
        return m

    def _accessible(self, start):
        return dfs(start, self.arcs).nodes

    def accessible(self):
        return self._accessible(self.start)

    @lru_cache(None)
    def trim(self):
        c = self.accessible() & self.reverse().accessible()
        m = FSA()
        for i in self.start & c:
            m.add_start(i)
        for i in self.stop & c:
            m.add_stop(i)
        for i,a,j in self.arcs():
            if i in c and j in c:
                m.add(i,a,j)
        return m

    def renumber(self):
        return self.rename(Integerizer())

    def rename(self, f):
        "Note: f is not bijective, states may split/merge."
        m = FSA()
        for i in self.start:
            m.add_start(f(i))
        for i in self.stop:
            m.add_stop(f(i))
        for i, a, j in self.arcs():
            m.add(f(i), a, f(j))
        return m

    def rename_apart(self, other):
        f = Integerizer()
        self = self.rename(lambda i: f((0, i)))
        other = other.rename(lambda i: f((1, i)))
        assert self.nodes.isdisjoint(other.nodes)
        return (self, other)

    def __mul__(self, other):
        m = FSA()
        self, other = self.rename_apart(other)
        m.start = self.start
        m.stop = other.stop
        for i in self.stop:
            for j in other.start:
                m.add(i,eps,j)
        for i,a,j in self.arcs():
            m.add(i,a,j)
        for i,a,j in other.arcs():
            m.add(i,a,j)
        return m

    def __add__(self, other):
        m = FSA()
        [self, other] = self.rename_apart(other)
        m.start = self.start | other.start
        m.stop = self.stop | other.stop
        for i,a,j in self.arcs():
            m.add(i,a,j)
        for i,a,j in other.arcs():
            m.add(i,a,j)
        return m

    def p(self):
        "self^+"
        m = FSA()
        m.start = set(self.start)
        m.stop = set(self.stop)
        for i,a,j in self.arcs():
            m.add(i,a,j)
        for i in self.stop:
            m.add_stop(i)
            for j in self.start:
                m.add(i, eps, j)
        return m

    def star(self):
        "self^*"
        return one + self.p()

#    def L(self, s):
#        assert s in self.nodes
#        return dfs({s}, self.arcs)

    @lru_cache(None)
    def epsremoval(self):

        eps_m = FSA()
        for i,a,j in self.arcs():
            if a == eps:
                eps_m.add(i,a,j)

        @lru_cache
        def eps_accessible(i):
            return eps_m._accessible({i})

        m = FSA()

        for i,a,j in self.arcs():
            if a == eps: continue
            m.add(i, a, j)
            for k in eps_accessible(j):
                m.add(i, a, k)

        for i in self.start:
            m.add_start(i)
            for k in eps_accessible(i):
                m.add_start(k)

        for i in self.stop:
            m.add_stop(i)
            for k in eps_accessible(i):
                m.add_stop(k)

        return m

    @lru_cache(None)
    def det(self):

        self = self.epsremoval()

        def powerarcs(Q):
            for a in self.syms:
                yield a, frozenset({j for i in Q for j in self.edges[i][a]})

        m = dfs([frozenset(self.start)], powerarcs)

        for powerstate in m.nodes:
            if powerstate & self.stop:
                m.add_stop(powerstate)

        return m

    def min_brzozowski(self):
        "Brzozowski's minimization algorithm"
        # https://en.wikipedia.org/wiki/DFA_minimization#Brzozowski's_algorithm

        # Proof of correctness:
        #
        # Let M' = M.r.d.r
        # Clearly,  [[M']] = [[M]]
        #
        # In M', there are no two states that can accept the same suffix
        # language because the reverse of M' is deterministic.
        #
        # The determinization of M' then creates powerstates, where every pair
        # of distinct powerstates R and S, there exists by construction at least
        # one state q of M' where q \in R and q \notin S. Such a q contributes
        # at least one word w \in [[q]] to the suffix language of q in [[R]]
        # that is not present in [[S]], since this word is unique to q (i.e., no
        # other state accepts it).  Thus, all pairs of states in M'.d are
        # distinguishable.
        #
        # Thus, after trimming of M'.d, we have a DFA with no indistinguishable
        # or unreachable states, which is must minimal.

        return self.reverse().det().reverse().det().trim()

    def min_fast(self):
        self = self.det().renumber()

        # calculate inverse of transition function (i.e., reverse arcs)
        inv = defaultdict(set)
        for i,a,j in self.arcs():
            inv[j,a].add(i)

        final = self.stop
        nonfinal = self.nodes - final

        P = [final, nonfinal]
        W = [final, nonfinal]

        while W:
            A = W.pop()
            for a in self.syms:
                X = {i for j in A for i in inv[j,a]}
                R = []
                for Y in P:
                    if X.isdisjoint(Y) or X >= Y:
                        R.append(Y)
                    else:
                        YX = Y & X
                        Y_X = Y - X
                        R.append(YX)
                        R.append(Y_X)
                        W.append(YX if len(YX) < len(Y_X) else Y_X)
                P = R

        # create new equivalence classes of states
        minstates = {}
        for i, qs in enumerate(P):
            #minstate = frozenset(qs)
            for q in qs:
                minstates[q] = i #minstate

        return self.rename(lambda i: minstates[i]).trim()

    def min_faster(self):
        self = self.det().renumber()

        # calculate inverse of transition function (i.e., reverse arcs)
        inv = defaultdict(set)
        for i,a,j in self.arcs():
            inv[j,a].add(i)

        final = self.stop
        nonfinal = self.nodes - final

        P = [final, nonfinal]
        W = [final, nonfinal]

        find = {i: block for block, elements in enumerate(P) for i in elements}

        while W:

            A = W.pop()
            for a in self.syms:

                X = {i for j in A for i in inv[j,a]}

                blocks = {find[i] for i in X}

                for block in blocks:
                    Y = P[block]

                    if X >= Y: continue

                    # TODO: use indexing to find nonempty (Y-X).
                    # Some notes:
                    #  - To be nonempty we need to find an element i* that is in Y
                    #    but not in X.  We already have an element i that is in Y
                    #    and X.
                    #  - We know that X and Y overlap (thanks to our indexing
                    #    trick).  Now, we want to filter out cases where X >= Y
                    #    because they do not need to be split.

                    YX = Y & X
                    Y_X = Y - X

                    # we will replace block with the intersection case (no
                    # need to update `find` index for YX elements)
                    P[block] = YX

                    new_block = len(P)
                    for i in Y_X:
                        find[i] = new_block

                    P.append(Y_X)
                    W.append(YX if len(YX) < len(Y_X) else Y_X)

        return self.rename(lambda i: find[i]).trim()

    min = lru_cache(None)(min_faster)

    def equal(self, other):
        return self.min()._dfa_isomorphism(other.min())

    def _dfa_isomorphism(self, other):
        "Find isomorphism between DFAs (if one exists)."

        # Requires that self and other are minimal DFAs

        # Theorem. If `self` and `other` are graphs with out-degree at most 1, then
        # the DFA works to determine whether G and H are isomorphic

        # A deterministic machine has exactly one start state

        # Two minimized DFAs are input
        # If the number of states is differs, these machines cannot be isomorphic
        if len(self.nodes) != len(other.nodes): return False
        if len(self.start) == 0: return len(other.start) == 0

        assert len(self.start) == 1 and len(other.start) == 1

        #self = self.renumber()
        #other = other.renumber()

        [p] = self.start
        [q] = other.start

        stack = [(p, q)]
        iso = {p: q}

        syms = self.syms | other.syms

        done = set()
        while stack:
            (p, q) = stack.pop()
            done.add((p,q))
            for a in syms:

                # presences of the arc has to be the same
                if (a in self.edges[p]) != (a in other.edges[q]):
                    return False

                if a not in self.edges[p]:
                    continue

                # machines are assumed deterministic
                [r] = self.edges[p][a]
                [s] = other.edges[q][a]

                if r in iso and iso[r] != s:
                    return False

                iso[r] = s
                if (r,s) not in done:
                    stack.append((r,s))

        #print('-----')
        #print(iso)
        #print(self)
        #print(self.rename(lambda x: iso.get(x, -1)))
        #print(other)
        #print('-----')

        return self.rename(iso.get) == other

    def to_regex(self):
        import numpy as np
        from semirings.regex import Symbol
        from semirings.kleene import kleene

        n = len(self.nodes)

        A = np.full((n,n), Symbol.zero)
        start = np.full(n, Symbol.zero)
        stop = np.full(n, Symbol.zero)

        ix = Integerizer(list(self.nodes))

        for i in self.nodes:
            for a, j in self.arcs(i):
                if a == eps:
                    A[ix(i),ix(j)] += Symbol.one
                else:
                    A[ix(i),ix(j)] += Symbol(a)

        for i in self.start:
            start[ix(i)] += Symbol.one

        for i in self.stop:
            stop[ix(i)] += Symbol.one

        return start @ kleene(A, Symbol) @ stop

    def __and__(self, other):
        "intersection"

        self = self.epsremoval().renumber()
        other = other.epsremoval().renumber()

        def product_arcs(Q):
            (q1, q2) = Q
            for a, j1 in self.arcs(q1):
                for j2 in other.edges[q2][a]:
                    yield a, (j1,j2)

        m = dfs({(q1, q2) for q1 in self.start for q2 in other.start},
                product_arcs)

        # final states
        for q1 in self.stop:
            for q2 in other.stop:
                m.add_stop((q1, q2))

        return m

    def add_sink(self, syms):
        "constructs a complete FSA"

        syms = set(syms)

        self = self.renumber()

        sink = len(self.nodes)
        for a in syms:
            self.add(sink, a, sink)

        for q in self.nodes:
            if q == sink: continue
            for a in syms - set(self.edges[q]):
                if a == eps: continue  # ignore epsilon
                self.add(q, a, sink)

        return self

    def __sub__(self, other):
        return self & other.invert(self.syms | other.syms)

    __or__ = __add__

    def __xor__(self, other):
        "Symmetric difference"
        return (self | other) - (self & other)

    def invert(self, syms):
        "create the complement of the machine"

        self = self.det().add_sink(syms)

        m = FSA()

        for i in self.nodes:
            for a, j in self.arcs(i):
                m.add(i, a, j)

        for q in self.start:
            m.add_start(q)

        for q in self.nodes - self.stop:
            m.add_stop(q)

        return m

    def __floordiv__(self, other):
        "left quotient self//other ≐ {y | ∃x ∈ other: x⋅y ∈ self}"

        # TODO: support NFA/epsilon arcs?
        self = self.epsremoval()
        other = other.epsremoval()

        # quotient arcs are very similar to product arcs except that the common
        # string is "erased" in the new machine.
        def quotient_arcs(Q):
            (q1, q2) = Q
            for a, j1 in self.arcs(q1):
                for j2 in other.edges[q2][a]:
                    yield eps, (j1, j2)

        m = dfs({(q1, q2) for q1 in self.start for q2 in other.start},
                quotient_arcs)

        # If we have managed to reach a final state of q2 then we can move into
        # the post-prefix set of states
        for (q1,q2) in set(m.nodes):
            if q2 in other.stop:
                m.add((q1, q2), eps, (q1,))

        # business as usual
        for q1 in self.nodes:
            for a, j1 in self.arcs(q1):
                m.add((q1,), a, (j1,))
        for q1 in self.stop:
            m.add_stop((q1,))

        return m

    def __truediv__(self, other):
        "right quotient self/other ≐ {x | ∃y ∈ other: x⋅y ∈ self}"
        return (self.reverse() // other.reverse()).reverse()   # reduce to left quotient on reversed languages

    def __lt__(self, other):
        "self ⊂ other"
        if self.equal(other): return False
        return (self & other).equal(self)

    def __le__(self, other):
        "self ⊆ other"
        return (self & other).equal(self)

    @classmethod
    def lift(cls, x):
        m = cls()
        m.add_start(0); m.add_stop(1); m.add(0,x,1)
        return m

    @classmethod
    def from_string(cls, xs):
        m = cls()
        m.add_start(xs[:0])
        for i in range(len(xs)):
            m.add(xs[:i], xs[i], xs[:i+1])
        m.add_stop(xs)
        return m

    @classmethod
    def from_strings(cls, Xs):
        m = cls()
        for xs in Xs:
            m.add_start(xs[:0])
            for i in range(len(xs)):
                m.add(xs[:i], xs[i], xs[:i+1])
            m.add_stop(xs)
        return m

    def __contains__(self, xs):
        d = self.det()
        [s] = d.start
        for x in xs:
            t = d.edges[s][x]
            if not t: break
            [s] = t
        return (s in d.stop)

    def merge(self, S, name=None):
        "merge states in `S` into a single state."
        if name is None: name = min(S)
        def f(s):
            return name if s in S else s
        m = FSA()
        for x in self.start:
            m.add_start(f(x))
        for x,a,y in self.arcs():
            m.add(f(x),a,f(y))
        for x in self.stop:
            m.add_stop(f(x))
        return m

eps = 'ε'

FSA.one = one = FSA()
one.add_start(0); one.add_stop(0)

FSA.zero = zero = FSA()


def test_visualization():
    a,b = map(FSA.lift, 'ab')
    ((a+b).star())._repr_html_()


def test_intersection():
    a,b,c = map(FSA.lift, 'abc')
    assert a.equal((a + a) & a)
    assert b.equal((a + b) & (b + c))
    assert b.equal((a.star() + b.star()) & (b + c))
    assert b.star().equal((a+b).star() & (b.star() + c.star()))
    assert ((a*b*c) & (a+b+c).star()).equal(a*b*c)


def test_complement():
    a,b,c = map(FSA.lift, 'abc')

    have = a - a
    want = zero
    assert have.equal(want)

    have = (a+b) - b
    want = a
    assert have.equal(want)

    have = (a+b) - (b+c)
    want = a
    assert have.equal(want)

    have = (a + b.star()) - (b+c)
    want = a + b * b * b.star() + one
    assert have.equal(want)

    # 1 b bb bbb bbbb bbbbb bbbbbb
    # 1   bb     bbbb       bbbbbb
    have = b.star() - (b*b).star()
    want = b * (b * b).star()
    assert have.equal(want)


def test_equality():
    a,b,c = map(FSA.lift, 'abc')

    assert a.equal(a)
    assert not a.equal(b)

    x = ((one + a) * a.star() * a.star() * a.star() * b)
    y = x.min()
    assert x.equal(y)
    assert y.equal(x)

    assert not (a * a.star()).equal(a.star())

    assert (one + a.star()).equal(a.star())
    assert (b * one + b * a.star()).equal(b * a.star())

    assert (a * a.star()).equal(a * a.star() + a * a * a.star())

    assert (a + b) != (b + a)
    assert (a + b + c).equal(b + a + c)

    assert ((a+b).star()).equal((a.star()*b.star()).star())

    assert (a+a).equal(a)   # idempotent

    x = a*a + a
    y = (a*a + a)*(a*a + a) + a

    assert not x.equal(y)


def test_min():
    a,b = map(FSA.lift, 'ab')
    z = zero
    for x1 in [a,b]:
        for x2 in [a,b]:
            for x3 in [a,b]:
                z += (x1 * x2 * x3).min()
    assert len(z.min().nodes) == 4
    print(z.min())

    assert len(((a + b)*(a + b)*(a + b)).min().nodes) == 4

    assert len(((one + a) * a.star() * a.star() * a.star() * b).min().nodes) == 2


def test_fsa_to_regex():
    from semirings.regex import Symbol
    a, b = map(FSA.lift, 'ab')
    m = b * a.star()
    m = m.min()

    A, B = map(Symbol, 'ab')
    assert m.to_regex() == B + B * A.star() * A


def test_quotient():
    a, b, c, d = map(FSA.lift, 'abcd')

    q = (a * b) // a
    assert q.equal(b)

    q = (a * b * c + c * d) // a
    assert q.equal(b * c + zero * c * d)

    q = (a * b * c * d + c * a * b * d) // (a * b)
    assert q.equal(c * d)

    q = (a * b) / b
    assert q.equal(a)

    q = ((a * b) / b) / a
    assert q.equal(one)

    q = (a * b) / (a * b * c)
    assert q.equal(zero)

    q = (a.star() * b) // a
    assert q.equal(a.star() * b)

    q = (a.star() * b) // a.star()
    print(q)
    print(q.min().renumber())
    assert q.equal(a.star() * b)   # is this correct?


    # L1\L2 = {y | ∃x ∈ L2: xy ∈ L1}
    # L1/L2 = {x | ∃y ∈ L2: xy ∈ L1}

    L1 = (a.star() * b)
    L2 = a.star()

    def checker(y): return (L2 * y) <= L1

    assert checker(a * b)
    assert checker(a.star() * b)
    assert not checker(c)


if __name__ == '__main__':
    from arsenal import testing_framework
    testing_framework(globals())



from . import base


class CutSets(base.Semiring):
    """Set of cut sets

    Implements: Martelli (1976)
    "A Gaussian Elimination Algorithm for the Enumeration of Cut Sets in a Graph"
    https://dl.acm.org/doi/abs/10.1145/321921.321928

    """
    def __init__(self, *sets):
        assert all(isinstance(x, (frozenset, set)) for x in sets)
        self.sets = frozenset(min_sets(map(frozenset, sets)))  # project onto minimal set
    def __repr__(self):
        return '{%s}' % (', '.join(repr(set(x) or {}) for x in self))
    def __hash__(self):
        return hash(self.sets)
    def __eq__(self, other):
        return self.sets == other.sets
    def __mul__(self, other):
        return CutSets(*(self.sets | other.sets))
    def __add__(self, other):
        return CutSets(*{
            (x | y)
            for x in self
            for y in other
        })
    def __iter__(self):
        return iter(self.sets)
    def star(self):
        return self.one #+ self
    @classmethod
    def lift(cls, e): return cls({e})


def min_sets(xs):
    """
    Add r to the collection.
    Returns True if r was subsumed by a more general rule in the program
    Returns False otherwise.
    """
    ys = []
    for x in xs:
        add(ys, x)
    return ys


def add(xs, r):
    # This implementation is pretty inefficient as it is based on linear scan
    rm = []
    for i, s in enumerate(xs):
        # branch r is already subsumed by branch s
        if s < r: return True
        # new branch subsumes existing branch, will be deleted in favor of the
        # more general branch.
        if r < s: rm.append(i)
    for i in reversed(sorted(rm)):
        del xs[i]
    xs.append(r)
    return False


CutSets.zero = CutSets(set())
CutSets.one = CutSets()



import numpy as np
from . import base


class Float(base.Semiring):
    def __init__(self):
        assert False, 'should never be called'

    zero = 0.0
    one = 1.0

    def approx_zero(x):
        return abs(x) < 1e-7   # XXX: YUCKY MAGIC NUMBER

    @classmethod
    def lift(cls, x, _):
        return x

    @staticmethod
    def star(x):
        if x == 1: return np.inf
        return 1/(1-x)

    @classmethod
    def multiplicity(cls,x,m):
        return x*m

    def metric(self, other):
        return abs(self - other)



import numpy as np
from numpy import log as _log, exp, isnan, log1p as _log1p, expm1
from arsenal import colors
from . import base


def log(x):
    if x <= 0:
        return float('-inf')
    return _log(x)


def log1p(x):
    if x <= -1:
        return float('-inf')
    return _log1p(x)


def log1pexp(x):
    """
    Numerically stable implementation of log(1+exp(x)) aka softmax(0,x).

    -log1pexp(-x) is log(sigmoid(x))

    Source:
    http://cran.r-project.org/web/packages/Rmpfr/vignettes/log1mexp-note.pdf
    """
    if x <= -37:
        return exp(x)
    elif -37 <= x <= 18:
        return log1p(exp(x))
    elif 18 < x <= 33.3:
        return x + exp(-x)
    else:
        return x


def log1mexp(x):
    """
    Numerically stable implementation of log(1-exp(x))

    Source:
    http://cran.r-project.org/web/packages/Rmpfr/vignettes/log1mexp-note.pdf
    """
    assert x <= 0
    a = abs(x)
    if 0 < a <= 0.693:
        return log(-expm1(-a))
    else:
        return log1p(-exp(-a))


class LogVal(base.Semiring):

    def __init__(self, pos, ell):
        self.pos = pos
        self.ell = ell

    @classmethod
    def lift(cls, x):
        return cls(x >= 0, log(abs(x)))

    def __eq__(self, other):
        return np.allclose(float(self), float(other))

    def __hash__(self):
        return hash(float(self))

    def __lt__(self, other):
        return float(self) < float(other)

    def is_zero(self):
        return self.ell <= float('-inf')

    def __float__(self):
#        if self.is_zero():
#            return 0.0
        if self.pos:
            return +float(exp(self.ell))
        else:
            return -float(exp(self.ell))

    def __mul__(self, b):
        c = LogVal.lift(0.0)
#        if self.is_zero() or b.is_zero():
#            return c
        c.pos = self.pos == b.pos
        c.ell = self.ell + b.ell
        return c

    def __truediv__(self, b):
        c = LogVal.lift(0.0)
        if self.is_zero():
            return c
#        if b.is_zero():
#            return c         # divide by zero: Should probably return NaN.
        c.pos = self.pos == b.pos
        c.ell = self.ell - b.ell
        return c

    __div__ = __truediv__

    def __sub__(self, b):
        c = LogVal.lift(0.0)
        c.ell = b.ell
        c.pos = not b.pos
        return self + c

    def __add__(self, b):
        c = LogVal.lift(0.0)
        a = self
        if a.ell < b.ell:
            a, b = b, a
        if a.is_zero():
            c.pos = b.pos
            c.ell = b.ell
            return c
        if b.is_zero():
            c.pos = a.pos
            c.ell = a.ell
            return c
        x = b.ell - a.ell
        assert not isnan(x)
#        assert x <= 0         # unnecessary assertion.
        if a.pos == 1 and b.pos == 1:
            c.pos = 1
            c.ell = a.ell + log1pexp(x)   # log(1+exp(x))
        elif a.pos == 1 and b.pos == 0:
            c.pos = 1
            c.ell = a.ell + log1mexp(x)   # log(1-exp(x))
        elif a.pos == 0 and b.pos == 1:
            c.pos = 0
            c.ell = a.ell + log1mexp(x)
        else:
            c.pos = 0
            c.ell = a.ell + log1pexp(x)
        return c

    def star(self):
        return LogVal.one / (LogVal.one - self)

    def multiplicity(self, m):
        return self.lift(m) * self

    def __repr__(self):
        x = float(self)
        sign = '+' if self.pos else colors.cyan % '-'
        return f'LogVal({x:.2g}={sign}exp({self.ell:.2g}))'

LogVal.zero = LogVal.lift(0.0)
LogVal.one = LogVal.lift(1.0)



import numpy as np
from arsenal import colors, assert_throws
from arsenal.iterextras import take
from semirings import (
    MinPlus, MinTimes, MaxPlus, MaxTimes, Float, CutSets, Boolean,
    Bottleneck, minmax, maxmin, LogVal, ConvexHull, Point,
    Lukasiewicz, Interval, LazySort, Dual, Bridge, Division,
    make_set, String, ThreeValuedLogic, VBridge, Wrapped,
    Why, Lineage, make_semiring
)
from semirings.regex import RegularLanguage
from semirings.wfsa import WFSA, EPSILON
from semirings.fsa import FSA


class WeightedGraph:

    def __init__(self, WeightType):
        self.N = set()
        self.WeightType = WeightType
        self.E = WeightType.chart()

    def __iter__(self):
        return iter(self.E)

    def incoming(self, j):  # TODO: use slice notation
        return {I: self.E[I,J] for I,J in self if J == j}

    def outgoing(self, i):  # TODO: use slice notation
        return {J: self.E[I,J] for I,J in self if I == i}

    def edge(self, i, j, w=None):
        self.N.add(i); self.N.add(j)
        self.E[i,j] += w if w is not None else self.WeightType.lift((i,j))

    # TODO: return another WeightedGraph that corresponds to the closure
    def __getitem__(self, item):
        i,j = item
        return self.E[i,j]

    def __setitem__(self, item, value):
        i,j = item
        self.N.add(i); self.N.add(j)
        self.E[i,j] = value
        return self

    def closure(self):
        # initialization
        A = self.E
        old = A.copy()
        for j in self.N:
            new = self.WeightType.chart()
            sjj = self.WeightType.star(old[j,j])
            for i in self.N:
                for k in self.N:
                    new[i,k] = old[i,k] + old[i,j] * sjj * old[j,k]
            old, new = new, old   # swap to repurpose space
        # post processing fix-up: add the identity matrix
        for i in self.N: old[i,i] += self.WeightType.one
        c = WeightedGraph(self.WeightType)
        c.E.update(old)
        for i,j in c.E: c.N.add(i); c.N.add(j)
        return c

    def _repr_svg_(self):
        return self.graphviz()._repr_image_svg_xml()

    def graphviz(self):
        from graphviz import Digraph

        name = Integerizer()

        g = Digraph(
            node_attr=dict(
                fontname='Monospace',
                fontsize='9',
                height='0', width='0',
                margin="0.055,0.042",
                penwidth='0.15',
                shape='box',
                style='rounded',
            ),
            edge_attr=dict(
                penwidth='0.5',
                arrowhead='vee',
                arrowsize='0.5',
                fontname='Monospace',
                fontsize='8'
            ),
        )

        def escape(x):
            if isinstance(x, frozenset): x = set(x) or {}
            if isinstance(x, float): x = f'{x:.2g}'
            x = str(x)
            # remove any ANSI color codes
            x = re.sub(r'\033\[.*?m(.*?)\033\[0m', lambda m: '`%s`' % m.group(1), x)
            return x

        for i,j in self.E:
            if self.E[i,j] == self.WeightType.zero: continue
            g.edge(str(name(i)), str(name(j)), label=escape(self.E[i,j]))

        for i in self.N:
            g.node(str(name(i)), label=escape(i))

        return g


def test_pow():
    assert Float.lift(0, None) ** 0 == 1
    assert Float.lift(1, None) ** 10 == 1
    assert Float.lift(2, None) ** 10 == 2**10
    assert Float.lift(0.5, None) ** 12 == 0.5**12

    assert (MaxPlus.lift(0.5, None) ** 12).score == 0.5 * 12
    assert (MaxPlus.lift(2, None) ** 0) == MaxPlus.one

    assert MaxTimes.lift(2, None) ** 5 == MaxTimes.lift(2 ** 5, None)


#def test_pareto():
#
#    def pareto(Y):
#        return np.array([y1
#                for y1 in Y
#                if not any(np.all(y2 < y1) for y2 in Y if y1 is not y2)
#               ])
#
#    points = np.random.uniform(size=(30,2))
#    P = pareto(points)
#
#    def show(points, c='r'):
#        if isinstance(points, Pareto): points = points.x
#        points = np.array(points)
#        P = pareto(points)
#        pl.scatter(points[:,0], points[:,1], alpha=0.5, c='b')
#        pl.scatter(P[:,0], P[:,1], c=c, alpha=0.5)
#
#    Pareto = make_semiring(
#        'Pareto',
#        lambda X,Y: pareto(list(X) + list(Y)),
#        lambda X,Y: pareto([x + y for x in X for y in Y]),
#        [],
#        [np.zeros(2)],
#    )

def test_uncertainty_sets():
    U = make_semiring(
        'uncertainty',
        plus = lambda X,Y: {x + y for x in X for y in Y},
        times = lambda X,Y: {x * y for x in X for y in Y},
        zero = {0},
        one = {1},
    )

    x = U({1})
    y = U({1})
    z = U({0,1})

    print(x, '+', y, '==', x + y)
    print(y, '*', z, '==', y * z)
    print(x, '*', z, '==', x * z)

    assert ((x + y) * z).x <= (x * z + y * z).x, [(x + y) * z, x * z + y * z]


def todo_wfsa():

    S = WFSA
    S.multiplicity = lambda x, m: x * WFSA.lift(EPSILON, m)

    a = S.lift('a', 1)
    b = S.lift('b', 1)
    c = S.lift('c', 1)

    members = [
        a,
        b,
        c,
        a.star() * b,
#        b.star(),
#        (a*b).star(),
        a * b.star() * c,
        (a + b) * c,
        (a * c + b * c),
#        (a.star() * b).star() * a.star(),
    ]

    # TODO: experimental support for hashing weighted regular languages
    if 0:
        have = set(); want = set()
        for A in members:
            for B in members:
                if hash(A) != hash(B):
                    have.add((A,B))
                if A.counterexample(B) is not None:
                    want.add((A,B))

        print('hash function:')
        precision = len(have & want) / len(have) if len(have) > 0 else 1
        recall = len(have & want) / len(want) if len(want) > 0 else 1
        print('  precision:', precision)
        print('  recall:   ', recall)

        if precision != 1:
            print()
            print(colors.light.red % 'false positives:')
            for a,b in have - want:
                #print(a.simple.to_wfsa())   # remove epsilons
                #print(b.simple.to_wfsa())

                print(a.min)
                print(b.min)

                print('hash:', hash(a), hash(b))
                print('counterexample:', a.counterexample(b))
                print()

        assert precision == 1

    check_axioms_samples(S,members)


def test_funky():
    # The semiring requires can be relaxed in many ways that still admit general
    # algorithms.  For example, we can relax the requirement that * is a monoid.
    # We can allow it to be a multi-arity operator with no structure other than
    # closure \forall x_1, \ldots, x_K \mathcal{X}: f(X_1...X_K) \in
    # \mathcal{X}.  But, we still want a kind of distributive property:
    #
    #   \forall k: \sum_{x_k} f(x_1, ..., x_k, ... x_K) = f(x_1, ..., (\sum_{x_k} x_k), ... x_K)
    #
    # Example from Gilea (2020) efficient outside computation
    # https://aclanthology.org/2020.cl-4.2/

    def F(x1, x2): return

    S = make_semiring(
        'Funky',
        lambda x,y: min(x, y),
        lambda x,y: x + np.exp(y),
        np.inf,
        1,   # ???
#        star = lambda x: 1 if x == 0 else K
#        pp = lambda x: repr(dict(x)),
#        hash = lambda x: hash(frozenset(x.items())),
#        multiplicity = multiplicity
    )

    members = list(map(S, np.linspace(-5,5,11)))
    print(members)

    check_axioms_samples(S,members,hash_=False,assoc=False,star=False)

    #print([A,B,C], A * (B + C), '~~', A * B + A * C)


def test_endomorphism():
    from collections import Counter

    S = make_semiring(
        'Endo',
        lambda f,g: lambda x: f(x) + g(x),
        lambda f,g: lambda x: f(g(x)),
        lambda _: 0,
        lambda x: x,
#        star = lambda x: 1 if x == 0 else K
#        pp = lambda x: repr(dict(x)),
#        hash = lambda x: hash(frozenset(x.items())),
#        multiplicity = multiplicity
    )

    # the tricky thing about the members is that the need to be endomorphisms
    # i.e., they have to preserve the monoid structure f(x) + f(y) = f(x + y).
    #
    # If the monoid is real addition, then f are affine functions.

    members = list(map(S, [
        lambda x: 2*x+1,
        lambda x: x*2,
        lambda x: x*3,
        lambda x: 4*x,
        lambda x: 1*x,
        lambda _: 1,
        lambda _: 3,
    ]))

    a,b,c,d,e,f,g = members

    def eval(f):
        return [f.x(x) for x in range(-5, 5)]

    assert eval((a+b)*c) == eval(a*c+b*c)

    print(eval(c*(a+b)))
    print(eval(c*a+c*b))

    assert eval(c*(a+b)) == eval(c*a+c*b)

#    for A in members:
#        for B in members:
#            for C in members:
#                check_axioms(S,A,B,C,star=False,hash_=False)


# XXX: not the "semiring needed" for the mst to work doesn't check out as a semiring!
def test_max_capacity():

    S = make_semiring(
        'MaxCapacity',
        max,
        min,
        -np.inf,
        np.inf,
        star = lambda x: np.inf
    )

    members = [
        S.zero,
        S.one,
        S(0.5),
        S(0.1),
        S(3),
        S(3.1415),
        S(2),
        S(0),
    ]

    for x in members:
        print(x)
        approx = S.star_approx(x, 100)
        analytical = S.star(x)
        print('  approx:    ', approx)
        print('  analytical:', analytical)
        assert approx == analytical, [approx, analytical]

    check_axioms_samples(S,members)


# XXX: work-in-progress
def test_semilinear():
    from collections import Counter

    def bag_union(x,y):
        z = Counter()
        for k,v in x.items(): z[k] += v
        for k,v in y.items(): z[k] += v
        return z

    def minkowski(x,y):
        z = Counter()
        for k1,v1 in x.items():
            for k2,v2 in y.items():
                assert len(k1) == len(k2)
                k3 = tuple(a + b for a,b in zip(k1,k2))
                z[k3] += v1*v2   # or should it be sum?
        return z


    from arsenal import Integerizer
    symbols = 'abcd'
    alphabet = Integerizer(list(symbols))
    emptyvector = (0,)*len(symbols)

    S = make_semiring(
        'Semilinear',
        bag_union,
        minkowski,
        Counter(),     # empty bag
        Counter({emptyvector: 0}),
#        star = lambda x: 1 if x == 0 else K
    )

    def onehot(symbol):
        x = [0]*len(symbols)
        x[alphabet(symbol)] = 1
        return S({tuple(x): 1})


    a,b,c,d = list(map(onehot, symbols))

    print(a,b,c,d)
    print(a + a*a + a*a*a + a*a*a + a*b*b + b*a*b + b*b*a)
    print(a + a*a + a*a*a + a*a*a + a*b*b + b*a*b + b*b*a)

    x = a
    print(x)
    approx = S.star_approx(x, 10)
    #analytical = S.star(x)
    print('  approx:    ', approx)
    #print('  analytical:', analytical)
    #if approx.score > 1000: approx = S.inf
    #assert approx == analytical, [approx, analytical]


def test_countK():

    # k-collapsed integers
    K = 5
    S = make_semiring(
        f'Count[{K}]',
        lambda a,b: min(K, a+b),
        lambda a,b: min(K, a*b),
        0,
        1,
        star = lambda x: 1 if x == 0 else K
    )

    members = list(map(S, range(K+1)))
    print(members)

    for x in members:
        print(x, S.star(x), S.star_fixpoint(x))

    check_axioms_samples(S,members)


def test_max_times():

    S = MaxTimes

    members = [
        S.zero,
        S.one,
        S(0.5),
        S(0.1),
        S(3),
        S(3.1415),
        S(2),
        S(0),
    ]

    for x in members:
        #print(x)
        approx = S.star_approx(x, 100)
        analytical = S.star(x)
        #print('  approx:    ', approx)
        #print('  analytical:', analytical)
        if approx.score > 1000: approx = S.inf
        assert approx == analytical, [approx, analytical]

    check_axioms_samples(S,members)


def test_ints():

    S = make_semiring(
        'Int',
        lambda a,b: a+b,
        lambda a,b: a*b,
        0,
        1,
        star = lambda x: 1 if x == 0 else np.inf,
        multiplicity = lambda x,m: x*m,
    )

    members = [
        S.zero,
        S.one,
        S(0),
        S(1),
        S(2),
        S(100),
#        S(-5),      # star fails (not well-defined)
#        S(np.inf),
    ]

    for x in members:
        print(x)
        approx = S.star_approx(x, 1000)
        analytical = S.star(x)
        print('  approx:    ', approx, [S.star_approx(x, k) for k in range(10)])
        print('  analytical:', analytical)
        if approx.x >= 1000: approx = S(np.inf)
        assert approx == analytical, [approx, analytical]

    check_axioms_samples(S,members)


def test_max_plus():

    S = MaxPlus

    members = [
        S.zero,
        S.one,
        S(0.5),
        S(0.1),
        S(3),
        S(3.1415),
        S(2),
        S(0),
        S(-3),
        S(+3),
        S(2),
        S(-1),
    ]

    for x in members:
        #print(x)
        approx = S.star_approx(x, 20_000)
        analytical = S.star(x)
        #print('  approx:    ', approx.score)
        #print('  analytical:', analytical.score)
        if approx.score > 1000: approx = S.inf
        assert approx == analytical, [approx.score, analytical.score]


    check_axioms_samples(S,members)


def test_security():

    # k-collapsed integers
    K = {-1,2,3,4}
    bot = min(K)
    top = max(K)
    S = make_semiring(
        'Security',
        min,
        max,
        top,
        bot,
        star = lambda _: bot,
    )

    members = list(map(S, K))

    for x in members:
        assert S.star(x) == S.star_fixpoint(x)

    check_axioms_samples(S,members)


def test_why_semiring():

    members = [
        Why.lift('a'),
        Why.lift('b'),
        Why.lift('c'),
    ]
    check_axioms_samples(Why,members)

    graph = WeightedGraph(Why)
    graph.edge(0,1)
    graph.edge(2,3)
    graph.edge(1,2)
    graph.edge(0,1.5)
    graph.edge(1.5,2)
    graph.edge(3,4)
    graph.edge(3,5)
    graph.edge(4,5)

    if 0:
        import pylab as pl
        graph.draw()
        pl.show()

    why = graph.closure()[0,5]   # we have a bridge at edge 2->3
    assert why == Why(frozenset({
        frozenset({(0, 1), (2, 3), (4, 5), (1, 2), (3, 4)}),
        frozenset({(0, 1.5), (2, 3), (4, 5), (3, 4), (1.5, 2)}),
        frozenset({(0, 1), (1, 2), (2, 3), (3, 5)}),
        frozenset({(0, 1.5), (2, 3), (1.5, 2), (3, 5)})
    }))


def test_lineage_semiring():

    members = [
        Lineage.lift('a'),
        Lineage.lift('b'),
        Lineage.lift('c'),
    ]

    check_axioms_samples(Lineage,members)

    graph = WeightedGraph(Lineage)
    graph.edge(0,1)
    graph.edge(2,3)
    graph.edge(1,2)
    graph.edge(0,1.5)
    graph.edge(1.5,2)
    graph.edge(3,4)
    graph.edge(3,5)
    graph.edge(4,5)

    if 0:
        import pylab as pl
        graph.draw()
        pl.show()

    c = graph.closure()

    # we use all the edges for 0->5 paths
    lineage = c[0,5]
    assert lineage == Lineage(frozenset(graph.E))

    # we only use four edges for 0->2 paths
    lineage = c[0,2]
    assert lineage == Lineage(frozenset({(0, 1), (1, 2), (1.5, 2), (0, 1.5)}))


def test_cut_sets():
    S = CutSets

    # minimality
    assert CutSets( {(2,3)}, {(1,3), (2,3)} ) == CutSets( {(2,3)} )

    a = CutSets( {(2,3)}, {(1,3), (2,4)} )
    b = CutSets( {(1,3), (2,3)}, {(1,3), (2,4)} )
    c = CutSets( {(1,3)} )
    assert a * b == CutSets({(2,3)}, {(1,3), (2,4)})
    assert a + b == CutSets({(1,3), (2,3)}, {(1,3), (2,4)})

    one = CutSets.one; zero = CutSets.zero
    assert a * one == a
    assert a * zero == zero
    assert a * (b + c) == a * b + a * c
    assert a * a == a
    assert a + a == a

    members = [
        CutSets({(2,3)}, {(1,3), (2,4)}),
        CutSets({(1,3), (2,3)}, {(1,3), (2,4)}),
        CutSets({(1,3)}),
        CutSets.zero,
        CutSets.one,
    ]
    check_axioms_samples(S,members)


def test_edge_bridge():
    S = Bridge
    members = [
        Bridge({(2,3), (1,3), (2,4)}),
        Bridge({(1,3), (2,3), (1,3), (2,4)}),
        Bridge({(1,3), (2,3)}),
        Bridge({(1,3), (2,4)}),
        Bridge.lift((1,3)),   # creates a singleton set
        Bridge.zero,
        Bridge.one,
    ]

    check_axioms_samples(S,members)

    graph = WeightedGraph(Bridge)
    graph.edge(0,1)
    graph.edge(2,3)
    graph.edge(1,2)
    graph.edge(0,1.5)
    graph.edge(1.5,2)
    graph.edge(3,4)
    graph.edge(3,5)
    graph.edge(4,5)

    if 0:
        import pylab as pl
        graph.draw()
        pl.show()

    bridges = graph.closure()[0,5]   # we have a bridge at edge 2->3
    assert bridges == Bridge({(2,3)})


def test_vertex_bridge():
    graph = WeightedGraph(VBridge)
    graph.edge(0,1)
    graph.edge(2,3)
    graph.edge(1,2)
    graph.edge(0,1.5)
    graph.edge(1.5,2)
    graph.edge(3,4)
    graph.edge(3,5)
    graph.edge(4,5)

    # clearly, removing 0 or 5 disconnects 0 from 5; but so will 2 or 3 (see picture above)
    b = graph.closure()[0,5]

    if 0:
        import pylab as pl
        graph.draw()
        pl.show()

    assert b == VBridge({0,2,3,5})


def test_division():
    S = Division
    members = [Division(x) for x in range(10)]
    check_axioms_samples(S,members)


def test_three_valued_logic():
    S = ThreeValuedLogic

    [f,u,t] = members = S.zero, S.unk, S.one

    assert f < u < t
    assert_equal(f + f, f)
    assert_equal(f + t, t)
    assert_equal(t + t, t)
    assert_equal(f + u, u)
    assert_equal(t + u, t)
    assert_equal(u + u, u)

    assert_equal(f * f, f)
    assert_equal(f * t, f)
    assert_equal(t * t, t)
    assert_equal(f * u, f)
    assert_equal(t * u, u)
    assert_equal(u * u, u)

    check_axioms_samples(S,members)


def test_sets():
    from arsenal.maths.combinatorics import powerset

    universe = frozenset('abc')
    S = make_set(universe, frozenset())

    members = [S(frozenset(x)) for x in powerset(universe)]

    check_axioms_samples(S,members)


def test_string():
    S = String
    members = [
        S.zero,
        S.one,
        S("a"),
        S("aa"),
        S("aaa"),
        S("b"),
        S("aaab"),
        S("abab"),
    ]

    assert S("aaa") + S("aab") == S("aa")
    assert S("aaa") + S("") == S("")

    check_axioms_samples(S,members,right_distrib=False)


def test_regular_sets():

    # We needed a version of RegularLanguage with | and & defined -- those ops
    # are currently only available on FSAs and FSAs have the wrong equality
    # operation.
    class RegularLanguageSet:
        def __init__(self, fsa):
            assert isinstance(fsa, FSA)
            self.fsa = fsa
        def __or__(self, other): return RegularLanguageSet(self.fsa | other.fsa)
        def __and__(self, other): return RegularLanguageSet(self.fsa & other.fsa)
        @classmethod
        def lift(cls, x): return cls(FSA.lift(x))
        def __eq__(self, other): return self.fsa.equal(other.fsa)
        def __hash__(self): return 0
        def __repr__(self): return f'RegularLanguageSet({self.fsa.min()})'
    RegularLanguageSet.zero = RegularLanguageSet(FSA.zero)
    RegularLanguageSet.one = RegularLanguageSet(FSA.one)

    a,b,c = map(FSA.lift, 'abc')

    S = make_set(RegularLanguageSet((a + b + c).star()), RegularLanguageSet.zero)

    members = [
        S.zero,
        S.one,
        S(RegularLanguageSet(a.star())),
        S(RegularLanguageSet((a+c).star() * b + b * (a+b).star())),
        S(RegularLanguageSet(b.star())),
        S(RegularLanguageSet((a + b).star())),
    ]

    check_axioms_samples(S,members)


def test_lazysort():

    def check(a, b, T=None):
        A = list(a.take(T))
        B = list(take(T, b))
        assert A == B, f'\n\n{A}\n{B}\n'

    a = LazySort(2,'a')
    b = LazySort(3,'b')
    c = LazySort(5,'c')

    z = LazySort.zero

    check(z.star(), LazySort.one)

    check(a * LazySort.one, a)
    check(LazySort.one * a, a)
    check(LazySort.one * LazySort.one, LazySort.one)

    check((a+b) * z, z)
    check(z * (a+b), z)

    check((a+b) * LazySort.one, a+b)
    check(LazySort.one * (a+b), a+b)

    check(a * LazySort.zero, [])
    check(LazySort.zero * a, [])

    check(a + LazySort.zero, a)
    check(LazySort.zero + a, a)

    check(a * b + a * c, a * (b + c))

    check(LazySort.multiplicity(a, 0), [])
    check(LazySort.multiplicity(a, 4), [a,a,a,a])
    check(LazySort.multiplicity(a, np.inf), [a,a,a,a,a], T=5)

    check(a.star(), [LazySort.one, a, a*a, a*a*a, a*a*a*a], T=5)


def test_bottleneck():

    one = Bottleneck.one
    zero = Bottleneck.zero

    a = Bottleneck(1)
    b = Bottleneck(2)
    c = Bottleneck(3)
    d = Bottleneck(4)

    assert a.star_approx(10) == a.star()
    assert a.star_fixpoint() == a.star()

    assert Bottleneck.sum([]) == zero
    assert Bottleneck.sum([a,b,c,d]) == a + b + c + d
    assert Bottleneck.product([]) == one
    assert Bottleneck.product([a,b,c,d]) == a * b * c * d

    assert (a * b + b * c + d * c) == c
    assert (a + b) == b
    assert (a * b) == a
    assert a * one == a
    assert a * zero == zero
    assert a + zero == a
    assert b + one == one
    print('ok')


def test_minmax():
    S = minmax(None, "")
    members = list(map(S, "abcd")) + [S.zero, S.one]
    check_axioms_samples(S,members)


def test_maxmin():
    S = maxmin(None, "")
    members = list(map(S, "abcd"))
    check_axioms_samples(S,members)


def test_convex_hull():
    check_axioms(
        ConvexHull,
        ConvexHull([Point(1,1.5), Point(2,1.75), Point(3,4)]),
        ConvexHull([Point(2,2)]),
        ConvexHull([Point(1,10)]),
        star=False,
    )


def test_star_operations():
    assert Float.star(1) == np.inf
    assert MinTimes(-1).star() == MinTimes(-np.inf)
    assert MinPlus(-1).star() == MinPlus(-np.inf)
    assert MaxPlus(1).star() == MaxPlus(np.inf)


# TODO:Intervals are not an exact semiring because they are sub-distributive
# rather than distributive.  Add specific tests for that!
def todo_interval():

    check_axioms(
        Interval,
        Interval(0.5,0.5),
        Interval(-1.1,+1.1),
        Interval(0.25,0.75),
    )

    with assert_throws(AssertionError):
        # fails distributivity
        check_axioms_samples(
            Interval,
            [Interval(1,1),
             Interval(-1,1),
             Interval(1,2)],
            star=False
        )

    assert Interval(0,1) | Interval(3,4) == Interval(0,4)

    assert Interval.tightest([]) == Interval.loose
    assert Interval.tightest([Interval(0,1)]) == Interval(0,1)
    assert Interval.tightest([Interval(0,1), Interval(-1,0.5)]) == Interval(0,0.5)

    assert Interval.union([]) == Interval.loose
    assert Interval.union([Interval(0,1)]) == Interval(0,1)
    assert Interval.union([Interval(0,1), Interval(-1,0.5)]) == Interval(-1,1)

    # TODO: should this be some special empty set?
    assert Interval(0,1) & Interval(3,4) == Interval(3,1)


def test_logval():

    samples = [
        LogVal.lift(.1),
        LogVal.lift(.2),
        LogVal.lift(.3),
        LogVal.lift(-0.1),
        LogVal.lift(0),
        LogVal.lift(10),
        LogVal.lift(23),
        LogVal.lift(-3.1e-5),
        LogVal.lift(-123.5),
        LogVal.lift(1e-100),
        LogVal.lift(1e+100),
    ]


    from semirings.logval import log1pexp, log1mexp
    xs = np.linspace(-40, 40, 100)
    ys = [log1pexp(x) for x in xs]
    zs = np.log(1+np.exp(xs))
    np.allclose(ys, zs)
    if 0:
        import pylab as pl
        pl.plot(xs, ys, c='b', alpha=0.5, label='log1pexp')
        pl.plot(xs, zs, c='r', alpha=0.5, linestyle=':', label='log(1+exp(x))')
        pl.show()

    xs = np.linspace(-40, 0, 100)
    ys = [log1mexp(x) for x in xs]
    zs = np.log(1-np.exp(xs))
    np.allclose(ys, zs)
    if 0:
        import pylab as pl
        pl.plot(xs, ys, c='b', alpha=0.5, label='log1mexp')
        pl.plot(xs, zs, c='r', alpha=0.5, linestyle=':', label='log(1-exp(x))')
        pl.show()

    assert LogVal.lift(.1) < LogVal.lift(.2)
    assert LogVal.lift(-.2) < LogVal.lift(-.1)
    assert LogVal.zero / LogVal.one == LogVal.zero

    for a in samples:
        for b in samples:
            for c in samples:
                check_axioms(LogVal, a, b, c)


def test_axioms():

    for S,samples in [
            (Lukasiewicz, [
                Lukasiewicz.zero,
                Lukasiewicz.one,
                Lukasiewicz(0.5),
                Lukasiewicz(0.25),
                Lukasiewicz(0.75),
            ]),
            (MinPlus, [
                MinPlus(float('+inf')),
                MinPlus.zero,
                MinPlus.one,
                MinPlus(-3),
                MinPlus(+3),
                MinPlus(2),
                MinPlus(-1)
            ]),
            (MaxPlus, [
                MaxPlus(float('-inf')),
                MaxPlus.zero,
                MaxPlus.one,
                MaxPlus(-3),
                MaxPlus(+3),
                MaxPlus(2),
                MaxPlus(-1),
            ]),
            (MinTimes, [
                MinTimes(float('+inf')),
                MinTimes.zero,
                MinTimes.one,
                MinTimes(+3),
                MinTimes(2),
            ]),
            (Dual, [
                Dual(0,1),
                Dual(-3,2),
                Dual(2,4),
                Dual(-1,-1),
            ]),
            (Float, [
                0, -3, 2,
                1, -1,
            ]),
            (Boolean, [
                Boolean(True),
                Boolean(False)
            ]),
            (RegularLanguage, [
                RegularLanguage.lift('a'),
                RegularLanguage.lift('b'),
                RegularLanguage.lift('c'),
                RegularLanguage.zero,
                RegularLanguage.one,
            ])
            #Symbol,   # equality operation of regex is insufficient
            #LazySort,
    ]:

        check_axioms_samples(S,samples)


def check_axioms_samples(S,samples,**kwargs):
    for A in samples:
        for B in samples:
            for C in samples:
                check_axioms(S,A,B,C,**kwargs)


def check_axioms(S,A,B,C,star=True,left_distrib=True,right_distrib=True,hash_=True,assoc=True,check_multiplicity=True):
    try:

        assert A == A
        if hash_:
            assert hash(A) == hash(A)
#            assert hash((A + B) * C) == hash(A * C + B * C)

        # Addition is a commutative monoid
        assert_equal((A + B) + C,
                     A + (B + C))
        assert_equal((A + B),
                     (B + A))
        assert_equal(A + S.zero,
                     A,
                     S.zero + A)

        # Multiplication is a monoid
        if assoc:
            assert_equal((A * B) * C, A * (B * C))
            assert_equal(A * S.one, A, S.one * A)

        # distributivity from the left and the right
        if left_distrib:  assert_equal(A * (B + C), A * B + A * C)
        if right_distrib: assert_equal((B + C) * A, B * A + C * A)

        # Annihilation
        assert_equal(A * S.zero, S.zero, S.zero * A)

        # Multiplicity operation
        if check_multiplicity:
            assert_equal(S.multiplicity(A,0) == S.zero)
            assert_equal(S.multiplicity(A,1) == A)
            assert_equal(S.multiplicity(A,2) == A + A)
            assert_equal(S.multiplicity(A,3) == A + A + A)
            #assert_equal(S.multiplicity(A,np.inf) == S.zero)

        if star:
            # Kleene star operation
            assert_equal(S.star(A),
                         S.one + S.star(A) * A,
                         S.one + A * S.star(A))

    except AssertionError as e:
        print()
        print(colors.light.cyan % S.__name__,A,B,C)
        # check semiring axioms for a subset of values
        raise e


def assert_equal(x, *ys):  # pragma: no cover
    if all(x == y for y in ys):
        #print(colors.ok, x, *ys)
        pass
    else:
        print(colors.bad, x, *ys)
        assert 0


if __name__ == '__main__':
    from arsenal import testing_framework
    testing_framework(globals())



from . import base


class MaxTimes(base.Semiring):

    def __init__(self, score, d=None):
        self.score = score
        self.d = d

    def __repr__(self):
        return f'MaxTimes({self.score}, {self.d})'

    def __eq__(self, other):
        return (isinstance(other, MaxTimes)
                and (self.score == other.score   # for infinity
                     or abs(self.score - other.score) < 1e-10))

    def __hash__(self):
        return hash(self.score)

    def __lt__(self, other):
        return isinstance(other, MaxTimes) and self.score < other.score

    def __add__(self, other):
        return max(self, other)

    def __mul__(self, other):
        if other is one: return self
        if self is one: return other
        if self is zero: return zero
        if other is zero: return zero
        return MaxTimes(self.score * other.score, (self.d, other.d))

    @classmethod
    def multiplicity(cls,x,m):
        if m > 0:
            return x
        else:
            return cls.zero

    def star(self):
        return self.one + self if self.score <= 1 else MaxTimes.inf


MaxTimes.zero = zero = MaxTimes(0, None)
MaxTimes.one = one = MaxTimes(1, ())
MaxTimes.inf = MaxTimes(float('+inf'), None)



from . import base


class MaxPlus(base.Semiring):

    def __init__(self, score, d=None):
        self.score = score
        self.d = d

    def __repr__(self):
        return f'MaxPlus({self.score}, {self.d})'

    def __eq__(self, other):
        return (isinstance(other, MaxPlus)
                and (self.score == other.score
                     or abs(self.score - other.score) < 1e-10))

    def __hash__(self):
        return hash(self.score)

    def __lt__(self, other):
        return isinstance(other, MaxPlus) and self.score < other.score

    def __add__(self, other):
        return max(self, other)

    def __mul__(self, other):
        return MaxPlus(self.score + other.score, [self.d, other.d])

    @classmethod
    def multiplicity(cls,x,m):
        if m > 0:
            return x
        else:
            return cls.zero

    def star(self):
        # a positive reward cycle goes to infinity,
        # everything else stays put x^* = max(0, x, x+x, x+x+x, ...)
        return (self.one + self) if self.score <= 0 else MaxPlus.inf


MaxPlus.zero = zero = MaxPlus(float('-inf'), None)
MaxPlus.one = one = MaxPlus(0.0, ())
MaxPlus.inf = MaxPlus(float('+inf'))



# TODO: Implement the semiring of square matrices


def kleene(A, semiring, reflexive=True):
    """
    Compute the matrix star
    """

    # initialization
    [N,_] = A.shape; zero = semiring.zero; one = semiring.one
    new = A.copy(); old = A.copy()
    for j in range(N):
        new[:,:] = zero
        sjj = semiring.star(old[j,j])
        for i in range(N):
            for k in range(N):
                # i ➙ j ⇝ j ➙ k
                new[i,k] = old[i,k] + old[i,j] * sjj * old[j,k]
        old, new = new, old   # swap to repurpose space
    if reflexive:  # post processing fix-up: add the identity matrix
        for i in range(N): old[i,i] += one
    return old



class Semiring:
    @classmethod
    def chart(cls):
        return Chart(cls.zero)
    zero = None
    one = None
#    def __lt__(self, other):
#        # Note: this arbitrary ordering is used by `Program.signature`
#        return hash(self) < hash(other)
    def __add__(self, other):
        raise NotImplementedError
    def __mul__(self, other):
        raise NotImplementedError
    @classmethod
    def lift(cls, *args):
        return cls(*args)
    @classmethod
    def multiplicity(cls, v, m):
        if m == 0: return cls.zero
        old = cls.zero
        while True:   # m might be infinite, in that case, run until we saturate
            new = v + old
            m -= 1
            if m == 0 or new == old: return new
            old = new
    @classmethod
    def multiple(cls, m):
        return cls.multiplicity(cls.one, m)
    def __pow__(self, n):
        assert n >= 0
        x = self
        if n == 0:       return self.one
        elif n % 2 == 0: return (x * x) ** (n // 2)
        else:            return x * (x ** (n - 1))
    def star_approx(self, T):
        # star(x) = 1 + x star(x)
        v = self.zero
        for _ in range(T):
            v = self.one + self * v
        return v
    def star_fixpoint(self):
        # star(x) = 1 + x star(x)
        prev = self.zero
        while True:
            curr = self.one + self * prev
            if prev == curr: return curr
            prev = curr
    star = star_fixpoint
    @classmethod
    def approx_zero(cls, x):
        return x == cls.zero
    @classmethod
    def sum(cls, xs):
        y = cls.zero
        for x in xs:
            y += x
        return y
    @classmethod
    def product(cls, xs):
        y = cls.one
        for x in xs:
            y *= x
        return y
    def __round__(self, precision):
        return self
    def metric(self, other):
        return (self != other)


class Chart(dict):

    def __init__(self, zero):
        self.zero = zero
        super(Chart, self).__init__()

    def __missing__(self, key):
        return self.zero

    def copy(self):
        x = Chart(self.zero)
        x.update(self)
        return x

#    def __add__(self, other):
#        assert isinstance(other, Chart), other
#        x = Chart(self.zero)
#        for k,v in self.items():
#            x[k] += v
#        for k,v in other.items():
#            x[k] += v
#        return x
#
#    def __sub__(self, other):
#        assert isinstance(other, Chart), other
#        x = Chart(self.zero)
#        for k,v in self.items():
#            x[k] += v
#        for k,v in other.items():
#            x[k] -= v
#        return x
#
#    def round(self, precision):
#        x = Chart(self.zero)
#        for k,v in self.items():
#            x[k] += round(v, precision)
#        return x


class Wrapped(Semiring):
    def __init__(self, x):
        self.x = x
    def __lt__(self, other): return self.x < other.x
    def __eq__(self, other): return self.x == other.x
    def __hash__(self):      return hash(self.x)
    def __repr__(self):      return f'{self.__class__.__name__}({self.x})'
    @classmethod
    def lift(cls, x):    return cls(x)


def make_semiring(name, plus, times, zero, one, star=None, pp=None, hash=hash, multiplicity=None):
    class SemiringWrapper(Semiring):
        def __init__(self, x, *args):
            self.x = x
        def __add__(self, other): return SemiringWrapper(plus(self.x, other.x))
        def __mul__(self, other):
            if isinstance(other, tuple):
                other = (SemiringWrapper.one if len(other) == 0 else other[0] * other[1:])
            return SemiringWrapper(times(self.x, other.x))
        def __lt__(self, other): return self.x < other.x
        def __eq__(self, other): return self.x == other.x
        def __hash__(self):      return hash(self.x)
        def __repr__(self):
            if pp is None:
                return f'{name}({self.x})'
            else:
                return pp(self.x)
        @classmethod
        def lift(cls, *args):    return cls(*args)

    if star is not None:
        SemiringWrapper.star = lambda self: SemiringWrapper(star(self.x))

    if multiplicity is not None:
        SemiringWrapper.multiplicity = lambda x,m: SemiringWrapper(multiplicity(x.x, m))

    SemiringWrapper.zero = SemiringWrapper(zero)
    SemiringWrapper.one = SemiringWrapper(one)
    SemiringWrapper.__name__ = name
    return SemiringWrapper



from semirings import Semiring
from arsenal.iterextras import merge_roundrobin, fair_product


from semirings.fsa import FSA
from functools import cached_property


class RegularLanguage(Semiring):

    @cached_property
    def fsa(self):
        raise NotImplementedError()

    def __eq__(self, other):
        return isinstance(other, RegularLanguage) and self.fsa.equal(other.fsa)

    def __hash__(self):
        return 0    # terrible hash function :-/

    def star(self):
        if self is NULL: return EPSILON
        if self is EPSILON: return EPSILON
        return Star(self)

    def __iter__(self):
        raise NotImplementedError()

    def __add__(self, other):
        if other is NULL: return self
        if self is NULL: return other
        return Union(self, other)

    def __mul__(self, other):
        if not isinstance(other, RegularLanguage): return other.__rmul__(self)
        # multiplication by zero
        if self is NULL or other is NULL: return NULL
        # multiplication by one.
        if self is EPSILON:  return other
        if other is EPSILON: return self
        return Concat(self, other)


class Union(RegularLanguage):
    def __init__(self, x, y):
        self.x = x
        self.y = y
        super().__init__()
    @cached_property
    def fsa(self):
        return self.x.fsa + self.y.fsa
    def __repr__(self): return f'{self.x} + {self.y}'
    def __iter__(self):
        yield from merge_roundrobin(self.x, self.y)


class Star(RegularLanguage):
    def __init__(self, x):
        self.x = x
        super().__init__()
    @cached_property
    def fsa(self):      return self.x.fsa.star()
    def __repr__(self): return f'({self.x})*'
    def __iter__(self):
        p = EPSILON
        while True:
            yield from p
            p = p * self.x


class Concat(RegularLanguage):
    def __init__(self, x, y):
        self.x = x
        self.y = y
        super().__init__()
    @cached_property
    def fsa(self):      return self.x.fsa * self.y.fsa
    def __repr__(self):
        x = self.x; y = self.y
        if isinstance(x, Union): x = f'({x})'
        if isinstance(y, Union): y = f'({y})'
        return f'{x}⋅{y}'
    def __iter__(self):
        for x,y in fair_product(self.x, self.y):
            yield x * y


class Symbol(RegularLanguage):
    def __init__(self, x):
        self.x = x
        super().__init__()
    @cached_property
    def fsa(self):      return FSA.lift(self.x)
    def __repr__(self): return f'{self.x}'
    def __iter__(self):
        yield self
#    def __call__(self, *args):
#        # check for NULL in args.
#        #if any(x == NULL for x in args): return NULL
#        return LabeledProduct(self, *args)


#class LabeledProduct(RegularLanguage):
#    def __init__(self, fn, *args):
#        self.fn = fn; self.args = args
#        super().__init__()
#    def __repr__(self):
#        if len(self.args) == 0:
#            return f'{self.fn}'
#        elif len(self.args) == 1:
#            #return f'{self.fn}⋅{self.args[0]}'  # to render as a string
#            return f'{self.fn}⟨{self.args[0]}⟩'
#        else:
#            return f'{self.fn}⟨{",".join(str(x) for x in self.args)}⟩'
#    def __iter__(self):
#        for a in fair_product(*self.args):
#            yield LabeledProduct(self.fn, *a)


class Zero(RegularLanguage):
    @cached_property
    def fsa(self):      return FSA.zero
    def __repr__(self): return '∅'
    def __iter__(self):
        if 0: yield
        return


class One(RegularLanguage):
    @cached_property
    def fsa(self):      return FSA.one
    def __repr__(self): return 'ε'
    def __iter__(self):
        yield self


EPSILON = RegularLanguage.one = One()
NULL = RegularLanguage.zero = Zero()
RegularLanguage.lift = Symbol


def evaluate(x, subst, semiring):
    def f(x):
        if   isinstance(x, Symbol): return subst[x]
        elif isinstance(x, Union):  return f(x.x) + f(x.y)
        elif isinstance(x, Concat): return f(x.x) * f(x.y)
        elif isinstance(x, Star):   return semiring.star(f(x.x))
        elif isinstance(x, Zero):   return semiring.zero
        elif isinstance(x, One):    return semiring.one
        else:                       raise ValueError(x)
    return f(x)


def test():
#    from arsenal.iterextras import take
#    from collections import Counter

#    def check(a, b, T=None):
#        assert a == b
#        A = list(take(T, a))
#        B = list(take(T, b))
#        assert Counter(A) == Counter(B), f'\n\n{A}\n{B}\n'

    a,b,c,d = map(Symbol, 'abcd')

    assert set(NULL) == set()
    assert set(EPSILON) == {EPSILON}
    assert set(NULL.star()) == {EPSILON}
    assert set(EPSILON.star()) == {EPSILON}

    assert a != b
    assert a == a

    assert (a*a)*a == a*(a*a)

    assert a*c != a*d, [a*c, a*d]

    ab = a*c + b*c + a*d + b*d
    assert len(set(ab)) == 4, list(ab)

    ab = (a + b) * (c + d)
    assert len(set(a + b)) == 2
    assert len(set(c + d)) == 2
    assert len(set(ab)) == 4, set(ab)

    assert a.star() == EPSILON + a * a.star()
    assert (a * b).star() == EPSILON + a * (b * a).star() * b
    assert (a + b).star() == a.star() * (a.star() * b).star() * a.star()

    assert hash(a.star()) == hash(a.star())
#    assert hash(a.star()) != hash(a)
#    assert hash(a + b) != hash(a * b)

#    for x in take(25, (a.star() + b.star()) * (c.star() + d.star())):
#        print(x)

    # Tests for labeled products
#    f = Symbol('f')
#    g = Symbol('g')
#    check(f(NULL), NULL)
#
#    check(f(a + b), f(a) + f(b))
#
#    check(g(f(a + b), c + d),
#          g(f(a), c + d) + g(f(b), c + d))
#
#    check(g(f(NULL + a + b), c + d),
#          g(f(a), c) + g(f(b), c) + g(f(a), d) + g(f(b), d))

    #check(EPSILON(a), a)

    print('test: pass')


if __name__ == '__main__':
    test()



"""
Implementation of weighted finite state automata over fields.

This implementation of equivalence and minimization follows Kiefer (2020) very closely


References

 - Stefan Kiefer (2020) "Notes on Equivalence and Minimization of Weighted Automata"
   arXiv:2009.01217v1 [cs.FL]


"""
import numpy as np
from arsenal import Integerizer, colors
from collections import defaultdict, Counter, deque
from functools import cached_property

from numpy import linalg
#from scipy import linalg

#cached_property = property


EPSILON = ε = "ε"


class WFSA:
    """
    Weighted finite state automata where weights are a field (e.g., real-valued).
    """
    def __init__(self):
        self.alphabet = set()
        self.states = set()
        self.delta = defaultdict(lambda: defaultdict(Counter))
        self.start = Counter()
        self.stop = Counter()

    def __repr__(self):
        return f"WFSA({self.dim} states)"

    def __hash__(self):
        return hash(self.simple)

    def __str__(self):
        output = []
        output.append('{')
        for p in self.states:
            output.append(f'  {p} \t\t({self.start[p]}, {self.stop[p]})')
            for a, q, w in self.arcs(p):
                output.append(f'    {a}: {q}\t[{w}]')
        output.append('}')
        return "\n".join(output)

    @property
    def dim(self):
        return len(self.states)

    def add_state(self, q):
        self.states.add(q)

    def add_arc(self, i, a, j, w):
        self.add_state(i)
        self.add_state(j)
        self.alphabet.add(a)
        self.delta[i][a][j] += w

    def add_I(self, q, w):
        self.add_state(q)
        self.start[q] += w

    def add_F(self, q, w):
        self.add_state(q)
        self.stop[q] += w

    def __call__(self, xs):
        "Evaluate the sequence `xs`."
        return self.simple(xs)

    @property
    def I(self):
        for q, w in self.start.items():
            if abs(w) > 1e-5:
                yield q, w

    @property
    def F(self):
        for q, w in self.stop.items():
            if abs(w) > 1e-5:
                yield q, w

    def arcs(self, i=None):
        if i is None:
            for i in self.delta:
                for a, T in self.delta[i].items():
                    for j, w in T.items():
                        if abs(w) > 1e-5:
                            yield i, a, j, w
        else:
            for a, T in self.delta[i].items():
                for j, w in T.items():
                    if abs(w) > 1e-5:
                        yield a, j, w

    def rename(self, f):
        "Note: If `f` is not bijective, states may merge."
        m = self.spawn()
        for i, w in self.I:
            m.add_I(f(i), w)
        for i, w in self.F:
            m.add_F(f(i), w)
        for i, a, j, w in self.arcs():
            m.add_arc(f(i), a, f(j), w)
        return m

    def rename_apart(self, other):
        f = Integerizer()
        return (self.rename(lambda i: f((0,i))), other.rename(lambda i: f((1,i))))

    @cached_property
    def renumber(self):
        return self.rename(Integerizer())

    def spawn(self, *, keep_init=False, keep_arcs=False, keep_stop=False):
        "returns a new FSA in the same semiring"
        m = WFSA()
        if keep_init:
            for q, w in self.I:
                m.add_I(q, w)
        if keep_arcs:
            for i,a,j,w in self.arcs():
                m.add_arc(i, a, j, w)
        if keep_stop:
            for q, w in self.F:
                m.add_F(q, w)
        return m

    @cached_property
    def reverse(self):
        "creates a reversed machine"
        # create the new machine
        R = self.spawn()
        # reverse each arc
        for i, a, j, w in self.arcs():
            R.add_arc(j, a, i, w)
        # reverse initial and final states
        for q, w in self.I:
            R.add_F(q, w)
        for q, w in self.F:
            R.add_I(q, w)
        return R

    def __add__(self, other):

        # non-essential optimizations
        if self is zero: return other
        if other is zero: return self

        self, other = self.rename_apart(other)
        U = self.spawn(keep_init=True, keep_arcs=True, keep_stop=True)
        # add arcs, initial and final states from argument
        for i, a, j, w in other.arcs():
            U.add_arc(i, a, j, w)
        for q, w in other.I:
            U.add_I(q, w)
        for q, w in other.F:
            U.add_F(q, w)
        return U

    def __sub__(self, other):
        self, other = self.rename_apart(other)
        U = self.spawn(keep_init=True, keep_arcs=True, keep_stop=True)
        # add arcs, initial and final states from argument
        for q, w in other.I:            U.add_I(q, -w)
        for i, a, j, w in other.arcs(): U.add_arc(i, a, j, w)
        for q, w in other.F:            U.add_F(q, w)
        return U

    def __mul__(self, other):
        if not isinstance(other, WFSA): return other.__rmul__(self)

        # non-essential optimizations
        if self is one: return other
        if other is one: return self
        if other is zero: return zero
        if self is zero: return zero

        self, other = self.rename_apart(other)
        C = self.spawn(keep_init=True, keep_arcs=True)
        # add arcs, initial and final states from argument
        for i, a, j, w in other.arcs():
            C.add_arc(i, a, j, w)
        for q, w in other.F:
            C.add_F(q, w)
        # connect the final states from `self` to initial states from `other`
        for i1, w1 in self.F:
            for i2, w2 in other.I:
                C.add_arc(i1, EPSILON, i2, w1 * w2)
        return C

    def star(self):
        return self.one + self.kleene_plus()

    def kleene_plus(self):
        "self^+"
        m = self.spawn(keep_init=True, keep_arcs=True, keep_stop=True)
        for i, w1 in self.F:
            for j, w2 in self.I:
                m.add_arc(i, EPSILON, j, w1*w2)
        return m

    def _repr_svg_(self):
#        return self.graphviz()._repr_svg_()
        return self.graphviz()._repr_image_svg_xml()

    def graphviz(self, fmt=lambda x: f'{round(x,3):g}', fmt_node=lambda x: ' '):
        from graphviz import Digraph
        g = Digraph(
            graph_attr=dict(rankdir='LR'),
            node_attr=dict(
                fontname='Monospace',
                fontsize='10',
                height='.05', width='.05',
                #margin="0.055,0.042"
                margin="0,0"
            ),
            edge_attr=dict(
                #arrowhead='vee',
                arrowsize='0.3',
                fontname='Monospace',
                fontsize='9'
            ),
        )
        f = Integerizer()
        for i,w in self.I:
            start = f'<start_{i}>'
            g.node(start, label='', shape='point', height='0', width='0')
            g.edge(start, str(f(i)), label=f'{fmt(w)}')
        for i in self.states:
            g.node(str(f(i)), label=str(fmt_node(i)), shape='circle')
        for i,w in self.F:
            stop = f'<stop_{i}>'
            g.node(stop, label='', shape='point', height='0', width='0')
            g.edge(str(f(i)), stop, label=f'{fmt(w)}')
        for i, a, j, w in sorted(self.arcs()):
            g.edge(str(f(i)), str(f(j)), label=f'{a}/{fmt(w)}')
        return g

    @cached_property
    def simple(self):
        self = self.renumber

        S = self.dim
        start = np.zeros(S)
        arcs = {a: np.zeros((S,S)) for a in self.alphabet}
        stop = np.zeros(S)

        for i, w in self.I:
            start[i] += w
        for i, a, j, w in self.arcs():
            arcs[a][i,j] += w
        for i, w in self.F:
            stop[i] += w

        if EPSILON in arcs:
            W = arcs.pop(EPSILON)   # remove it
            E = linalg.inv(np.eye(self.dim) - W)
            for a in arcs:
                arcs[a] = arcs[a] @ E
            start = start @ E

        return Simple(start, arcs, stop)

    def __eq__(self, other):
        return self.simple == other.simple

    def counterexample(self, other):
        return self.simple.counterexample(other.simple)

    @cached_property
    def min(self):
        return self.simple.min.to_wfsa()

    def multiplicity(self, m):
        return WFSA.lift(EPSILON, m) * self

#    def push(self):
#        """
#        Weight pushing
#        """
#        V = self.backward()
#        p = self.spawn()
#        for i in self.states:
#            p.add_I(i, self.start[i] * V[i])
#            p.add_F(i, 1/V[i] * self.stop[i])
#            for a, j, w in self.arcs(i):
#                p.add_arc(i, a, j, 1/V[i] * w * V[j])
#        return p
#
#    def backward(self):
#        b = self.R.chart()
#        K = self.matrix_star()
#        for p in self.states:
#            for q in self.states:
#                b[p] += K[p, q] * self.stop[q]
#        return b

    @classmethod
    def lift(cls, x, w):
        m = cls()
        m.add_I(0, 1)
        m.add_arc(0, x, 1, w)
        m.add_F(1, 1)
        return m

    @classmethod
    def from_string(cls, xs):
        m = cls()
        m.add_I(xs[:0], 1)
        for i in range(len(xs)):
            m.add_arc(xs[:i], xs[i], xs[:i+1], 1)
        m.add_F(xs, 1)
        return m

    @classmethod
    def from_strings(cls, Xs):
        m = cls()
        for xs in Xs:
            m.add_I(xs[:0])
            for i in range(len(xs)):
                m.add_arc(xs[:i], xs[i], xs[:i+1], 1)
            m.add_F(xs)
        return m


WFSA.zero = zero = WFSA()
WFSA.one = one = WFSA.lift(EPSILON, 1)


class Simple:
    def __init__(self, start, arcs, stop):
        assert EPSILON not in arcs
        self.start = start
        self.arcs = arcs
        self.stop = stop
        [self.dim] = start.shape

    def __call__(self, xs):
        forward = self.start.copy()
        for x in xs:
            if x not in self.arcs: return 0
            forward = forward @ self.arcs[x]
        return forward @ self.stop

    def __eq__(self, other):
#        return (self is other) or hash(self) == hash(other) and self.counterexample(other) is None
        return self.counterexample(other) is None

    def __hash__(self):
#        return self._hash
        return 0

    # TODO: not working yet :-(
    @cached_property
    def _hash(self):
        # Assign a random value in to each symbols of the alphabet, then compute
        # the weight of all paths.
        #
        # The idea for this hash function is related to this randomized
        # algorithm for equivalence testing: https://arxiv.org/abs/1302.2818
        #
        # TODO: We might be able to skip the matrix inverse! Rather than summing
        # over all paths, we can basically just sample some paths.  That's more
        # similar to what is done in that paper.
        self = self.min
        # this gives us a weighted graph where (i,a:w,j) ==> (i,a*w,j)
        n = self.dim
        W = np.zeros((n,n))
        for a,M in self.arcs.items():
            W += hash(a) * M
        z = self.start @ linalg.solve(np.eye(n) - W, self.stop)
        return hash(int(np.round(z)))

    def counterexample(self, B):

        alphabet = set(self.arcs) | set(B.arcs)

        va = self.start @ self.stop
        vb = B.start @ B.stop
        if not approx_equal(va, vb):
            return ((), va, vb)

        eta = np.hstack([self.stop, B.stop])
        if approx_equal(eta, 0):
            return  # both empty; thus, equivalent.

        worklist = deque()
        worklist.append(((), self.stop, B.stop))
        basis = [eta]

        while worklist:

            (w, VA, VB) = worklist.pop()

            for a in alphabet:

                ua = self.arcs[a] @ VA if a in self.arcs else 0*VA
                ub = B.arcs[a] @ VB if a in B.arcs else 0*VB

                w1 = (a, w)

                # counterexample?
                va = self.start @ ua
                vb = B.start @ ub
                if not approx_equal(va, vb):
                    return (w1, va, vb)   # yup!

                u = np.hstack([ua, ub])
                q = proj(u, basis)

                # add to the basis if it's not redundant; updates to the basis
                # go on the worklist.
                if not approx_equal(u - q, u):
                    worklist.append((w1, ua, ub))
                    basis.append(q)

    @cached_property
    def min(self):
        # Proposition 3.4: Forward (backward) conjugates are forward (backward) minimal.
        # Proposition 3.5: The backward (forward) conjugate of a forward (backward) minimal automaton is minimal.
        return self.forward_conjugate().backward_conjugate()

    def __repr__(self):
        return f'<Simple states={len(self.start)}, syms={len(self.arcs)}>'

    def forward_conjugate(self):
        # The forward basis F is an n' x n where n is the set of old states, and n' the set of new states
        #
        # we want to solve
        #
        #   old_start = new_start    F
        #       n          n'     (n' x n)
        #
        # There are n >= n' constraints, but they aren't linearly independent so
        # the system has a unique solution.
        #
        #    old_start @ F.T = new_start @ F @ F.T
        #                                 (n' x n) (n x n')
        #        n
        #
        #    old_start @ F.T @ inv(F @ F.T) = new_start
        F = self.forward_basis()

        #P = F.T @ linalg.inv(F @ F.T)
        P = linalg.pinv(F)

        #start = self.start @ P
        #assert np.allclose(self.start, start @ F)

        # We also need to solve for our new transition function
        #                   F M = M' F
        #               F M F.T = M' (F F.T)
        #    F M F.T inv(F F.T) = M' (F F.T) inv(F F.T)
        return Simple(
            start = self.start @ P,
            arcs = {a: F @ M @ P for a, M in self.arcs.items()},   # apply change of basis
            stop = F @ self.stop,
        )

    @cached_property
    def reverse(self):
        return Simple(
            start = self.stop,
            arcs = {a: M.T for a, M in self.arcs.items()},
            stop = self.start,
        )

    def forward_basis(self):
        worklist = [self.start]
        basis = [self.start]
        while worklist:
            V = worklist.pop()
            for a in self.arcs:
                u = V @ self.arcs[a]
                q = proj(u, basis)
                if not approx_equal(u - q, u):
                    worklist.append(u)
                    basis.append(q)
        return np.array(basis)

    def backward_conjugate(self):
        return self.reverse.forward_conjugate()

    def to_wfsa(self):
        m = WFSA()
        for i in range(self.dim):
            m.add_I(i, self.start[i])
        for a in self.arcs:
            for i in range(self.dim):
                for j in range(self.dim):
                    m.add_arc(i,a,j,self.arcs[a][i,j])
        for i in range(self.dim):
            m.add_F(i, self.stop[i])
        return m


def approx_equal(x,y):
    return np.allclose(x,y)


def proj(u, Q):
    # simple implementation of the Gram–Schmidt process
    # https://en.wikipedia.org/wiki/Gram%E2%80%93Schmidt_process
    for q in Q:
        u = u - ((q @ u) / (q @ q)) * q
    return u


def test_min():

    a = WFSA.lift('a', 1)
    b = WFSA.lift('b', 1)
    c = WFSA.lift('c', 1)

    print(colors.yellow % 'Example 1')
    M = one + a + a * b + a * b * c
    m = M.min
    assert m.dim == 4, m.dim

    print(colors.yellow % 'Example 2')
    z = zero
    for x1 in [a,b]:
        for x2 in [a,b]:
            for x3 in [a,b]:
                z += (x1 * x2 * x3)
    assert z.min.dim == 4

    print(colors.yellow % 'Example 3')
    assert ((a + b)*(a + b)*(a + b)).min.dim == 4

    print(colors.yellow % 'Example 4')
    M = ((one + a * a.star()) * b)
    m = M.min
    #print(m)
    #print(m.to_wfsa().push())

    assert m.dim == 2, m.dim


def test_equivalence():

    a = WFSA.lift('a', 1)
    b = WFSA.lift('b', 1)
    c = WFSA.lift('c', 1)


    #___________________________________________________________________________
    #
    print(colors.yellow % 'Example 0')
    A = a.star()
    B = a.star()

    w = A.counterexample(B)
    #print('counterexample?', w)
    assert w is None

    assert np.allclose(A('a'), 1)
    assert np.allclose(A('a'), 1)

    assert np.allclose(A('b'), 0)
    assert np.allclose(B('b'), 0)

    #___________________________________________________________________________
    #
    print(colors.yellow % 'Example 1')
    A = a * (b + c)
    B = a * b + a * c

    w = A.counterexample(B)
    #print('counterexample?', w)
    assert w is None

    assert np.allclose(A('ab'), 1), A('ab')
    assert np.allclose(A('ac'), 1)
    assert np.allclose(A('ab'), 1)
    assert np.allclose(B('ac'), 1)

    #___________________________________________________________________________
    #
    print(colors.yellow % 'Example 2')
    A = a * (b + c + c)
    B = a * b + a * c

    w = A.counterexample(B)
    #print('counterexample?', w)
    assert w is not None

    #___________________________________________________________________________
    #
    print(colors.yellow % 'Example 3')
    A = a * (b + c + c)
    B = a * b + a * c + a * c

    w = A.counterexample(B)
    #print('counterexample?', w)
    assert w is None

    #___________________________________________________________________________
    #
    print(colors.yellow % 'Example 4')
    A = a * b.star() * c.star()
    B = a * (one + b * b.star()) * c.star()

    w = A.counterexample(B)
    #print('counterexample?', w)
    assert w is None

    #___________________________________________________________________________
    #
    print(colors.yellow % 'Example 5')
    A = a.star()
    B = one + a * a.star()

    w = A.counterexample(B)
    #print('counterexample?', w)
    assert w is None

    #___________________________________________________________________________
    #
    print(colors.yellow % 'Example 6')

    A = ((one + a * a.star()) * b)
    B = (a.star() * b)
    assert A.counterexample(B) is None

    w = A.counterexample(B)
    #print('counterexample?', w)
    assert w is None

    #___________________________________________________________________________
    #
    print(colors.yellow % 'Example 7')
    A = (a * b).star()
    B = one + a * (b * a).star() * b

    w = A.counterexample(B)
    #print('counterexample?', w)
    assert w is None

    #___________________________________________________________________________
    #
    print(colors.yellow % 'Example 8')
    a = WFSA.lift('a', .1)
    b = WFSA.lift('b', .2)
    c = WFSA.lift('c', .3)

    A = (a + b).star()
    B = (a.star() * b).star() * a.star()

    w = A.counterexample(B)
    #print('counterexample?', w)
    assert w is None



if __name__ == '__main__':
    from arsenal import testing_framework
    testing_framework(globals())



import numpy as np
from semirings import Semiring


class Sample(Semiring):

    def __init__(self, w, d):
        self.w = w
        self.d = d

    def __iter__(self):
        while True:
            X = -np.log(np.random.uniform(0,1)) / self.w
            yield (X, self.d)

    def __add__(self, other):
        if other is zero: return self
        if self is zero: return other
        return Sum(self, other)

    def __mul__(self, other):
        if other is one: return self
        if self is one: return other
        if other is zero: return zero
        if self is zero: return zero
        return Prod(self, other)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.w}, {self.d})'


class Sum(Sample):
    def __init__(self, x, y):
        self.x = x; self.y = y
        super().__init__(x.w + y.w, d=[x,y])
    def __iter__(self):
        while True:
            for x,y in zip(self.x, self.y):
                yield min(x, y)


class Prod(Sample):
    def __init__(self, x, y):
        self.x = x; self.y = y
        super().__init__(x.w * y.w, d=[x,y])
    def __iter__(self):
        while True:
            for x,y in zip(self.x, self.y):
                xs,xd = x
                _,yd = y

                # We have two random variates
                #  X = -log(Ux)/Wx ~ Expon(Wx)
                #  Y = -log(Uy)/Wy ~ Expon(Wy)
                # We want to construct a new one using existing randomness
                #  Z = -log(U)/(Wx*Wy)

                # A simple manipulation shows that dividing by Wy works!
                #   X/Wy = (-log(Ux)/Wx)/Wy ~ Z

                Z = xs / self.y.w

                # We could, instead, generate a fresh random variate
                #Z = -np.log(np.random.uniform(0,1)) / (self.x.w * self.y.w)

                yield [Z, [xd, yd]]


zero = Sample(0.0, None)
one = Sample(1.0, ())

Sample.one = one
Sample.zero = zero



from setuptools import setup

setup(name='semirings',
      version='0.1.1',
      description='Semirings are a powerful abstraction for dynamic programming.',
      author='Tim Vieira',
      project_url = 'https://github.com/timvieira/semirings',
      install_requires = [
          'numpy',
          'scipy',       # for convex hull support
          'graphviz',
          'pytest',
          'arsenal',     # https://github.com/timvieira/arsenal
      ],
      long_description=open('README.md').read(),
      packages = ['semirings'],
)



from . import base


# WARNING: Intervals are not a semiring.
def make_interval(*, bot, zero, top):

    class Interval(base.Semiring):
        def __init__(self, lo, hi):
            self.lo = lo
            self.hi = hi
            self.empty = lo > hi

        def __repr__(self):
            if self.lo > self.hi: return '∅'
            if self.lo == self.hi: return f'[{self.lo}]'
            return f'[{self.lo},{self.hi}]'

        def __hash__(self):
            return hash((self.lo, self.hi))

        def __eq__(self, other):
            return self.empty == other.empty and self.lo == other.lo and self.hi == other.hi

#        def __lt__(self, other):
#            "definitely less than"
#            # [lo, hi] < [lo', hi']
#            #return self.hi < other.lo
#            raise NotImplementedError()

        def __add__(self, other):
            if self.empty: return empty
            if other.empty: return empty
            return Interval(self.lo + other.lo,
                            self.hi + other.hi)
#            a = self.lo + other.lo
#            b = self.lo + other.hi
#            c = self.hi + other.lo
#            d = self.hi + other.hi
#            return Interval(min([a,b,c,d]),
#                            max([a,b,c,d]))

        def __mul__(self, other):
            if self.empty: return empty
            if other.empty: return empty
            a = self.lo * other.lo
            b = self.lo * other.hi
            c = self.hi * other.lo
            d = self.hi * other.hi
            return Interval(min([a,b,c,d]),
                            max([a,b,c,d]))

        def inv(self):
            if self.lo > self.hi:                  return self   # no-op on an empty interval
            elif self.hi == 0:                     return Interval(bot, 1/self.lo)
            elif self.lo == 0:                     return Interval(1/self.hi, top)
            elif (self.lo <= zero <= self.hi):     return Interval.loose
            else:                                  return Interval(1/self.hi, 1/self.lo)  # if zero is not in interval

        def __truediv__(self, other):
            return self * other.inv()

        def __neg__(self):
            if self.empty: return empty
            # -self swaps lo and hi because it *reflects* across 0.
            return Interval(-self.hi, -self.lo)

        def __sub__(self, other):
            return self + -other

        def __or__(self, other):
            "(lossy) union of intervals"
            return Interval(min(self.lo, other.lo), max(self.hi, other.hi))

        def __and__(self, other):
            "intersection of intervals"
            return Interval(max(self.lo, other.lo), min(self.hi, other.hi))

        def __contains__(self, x):
            return not self.empty and self.lo <= x.lo <= x.hi <= self.hi

        @classmethod
        def sum(cls, xs):
            y = cls.zero
            for x in xs:
                y += x
            return y

        @classmethod
        def tightest(cls, xs):
            xs = list(xs)
            if len(xs) == 0: return cls.loose
            return cls(max(x.lo for x in xs), min(x.hi for x in xs))

        @classmethod
        def union(cls, xs):
            xs = list(xs)
            if len(xs) == 0: return cls.loose
            return cls(min(x.lo for x in xs), max(x.hi for x in xs))

        def star(self):
            return Interval(1,1)/(Interval(1,1) - self)

        def multiplicity(self, m):
            return Interval(m, m) * self

    empty = Interval(top, bot)
    Interval.zero = Interval(zero, zero)
    Interval.loose = Interval(bot, top)
    return Interval


Interval = make_interval(bot=float('-inf'), zero=0, top=float('+inf'))
Interval.one = Interval(1,1)



from . import base


class MinTimes(base.Semiring):

    def __init__(self, cost, d=None):
        self.cost = cost
        self.d = d

    def __repr__(self):
        return f'MinTimes({self.cost}, {self.d})'

    def __eq__(self, other):
        return isinstance(other, MinTimes) and self.cost == other.cost

    def __hash__(self):
        return hash(self.cost)

    def __lt__(self, other):
        return isinstance(other, MinTimes) and self.cost < other.cost

    def __add__(self, other):
        return min(self, other)

    def __mul__(self, other):
        if other is one: return self
        if self is one: return other
        if self is zero: return zero
        if other is zero: return zero
        return MinTimes(self.cost * other.cost, (self.d, other.d))

    @classmethod
    def multiplicity(cls,x,m):
        if m > 0:
            return x
        else:
            return cls.zero

    def star(self):
        if self.cost < 0: return MinTimes(float('-inf'), None)
        return self.one + self


MinTimes.zero = zero = MinTimes(float('inf'), None)
MinTimes.one = one = MinTimes(1, ())



from semirings.base import Semiring, Wrapped


class Why(Wrapped):
    """
    Set of combinations of tuples (annotations) needed for a tuple to exist in
    the answer to a query (or derive a given item).
    """
    def __add__(self, other): return Why(self.x | other.x)
    def __mul__(self, other): return Why(frozenset(x | y
                                                   for x in self.x
                                                   for y in other.x))
    @classmethod
    def lift(cls, x): return  cls(frozenset([frozenset([x])]))

Why.zero = Why(frozenset())
Why.one = Why(frozenset([frozenset([])]))


class Lineage(Wrapped):
    "Set of reachable annotations"

    def __add__(self, other):
        if self == self.zero: return other
        if other == self.zero: return self
        return Lineage(self.x | other.x)
    def __mul__(self, other):
        if self == self.zero: return self.zero
        if other == self.zero: return self.zero
        return Lineage(self.x | other.x)
    @classmethod
    def lift(cls, x): return  cls(frozenset([x]))

Lineage.zero = Lineage(None)
Lineage.one = Lineage(frozenset())


class Bridge(Semiring):
    "Edge bridge"
    def __init__(self, xs):  # set of edges
        self.xs = frozenset(xs) if xs is not None else xs
    def __add__(self, other):
        if self == self.zero: return other
        if other == self.zero: return self
        return Bridge(self.xs & other.xs)
    def __mul__(self, other):
        if self == self.zero: return self.zero
        if other == self.zero: return self.zero
        return Bridge(self.xs | other.xs)
    def __eq__(self, other):
        return self.xs == other.xs
    def __hash__(self):
        return hash(self.xs)
    @classmethod
    def lift(cls, x):
        return cls({x})
    def __repr__(self):
        return repr(self.xs)
Bridge.zero = Bridge(None)
Bridge.one = Bridge(set())


class VBridge(Bridge):
    "Vertex bridge"
    @classmethod
    def lift(cls, x):
        (i,j) = x
        return cls({i,j})


class Lukasiewicz(Wrapped):
    """The Łukasiewicz semiring: the closed interval [0,1] with addition given by
    max(a,b) and multiplication given by max(0, a + b - 1) appears in
    multi-valued logic.
    """
    def __add__(self, other): return max(self, other)
    def __mul__(self, other): return Lukasiewicz(max(0, self.x + other.x - 1))
Lukasiewicz.zero = Lukasiewicz(0)
Lukasiewicz.one = Lukasiewicz(1)


def make_set(universe, emptyset):
    class Set(Wrapped):
        """
        Semiring of sets
        """
        def __add__(self, other): return Set(self.x | other.x)
        def __mul__(self, other): return Set(self.x & other.x)
    Set.zero = Set(emptyset)
    Set.one = Set(universe)
    return Set


from math import gcd
lcm = lambda x,y: abs(x * y) // gcd(x,y) if x * y != 0 else 0
class Division(Wrapped):
    def __add__(self, other): return Division(gcd(self.x, other.x))
    def __mul__(self, other): return Division(lcm(self.x, other.x))
Division.zero = Division(0)
Division.one = Division(1)


#class WrappedBackpointers(Semiring):
#    def __init__(self, x, d=None):
#        self.x = x
#        self.d = d   # ignored in equality test
#    def __lt__(self, other): return self.x < other.x
#    def __eq__(self, other): return self.x == other.x
#    def __hash__(self):      return hash(self.x)
#    def __repr__(self):      return f'{self.__class__.__name__}({self.x})'
#    @classmethod
#    def lift(cls, *args):    return cls(*args)


def minmax(zero, one):
    class minmax(Wrapped):
        def __add__(self, other):
            if self == zero: return other
            if other == zero: return self
            return min(self, other)
        def __mul__(self, other):
            if self == zero: return zero
            if other == zero: return zero
            if self == one: return other
            if other == one: return self
            return max(self, other)
        def multiplicity(self, m):
            return self if m > 0 else self.zero
    minmax.zero = zero = minmax(zero)
    minmax.one = one = minmax(one)
    return minmax


def maxmin(zero, one):

    class maxmin(Wrapped):
        def __add__(self, other):
            if self == zero: return other
            if other == zero: return self
            return max(self, other)
        def __mul__(self, other):
            if self == zero: return zero
            if other == zero: return zero
            if self == one: return other
            if other == one: return self
            return min(self, other)
        def multiplicity(self, m): return self if m > 0 else self.zero

    maxmin.zero = zero = maxmin(zero)
    maxmin.one = one = maxmin(one)
    return maxmin


Bottleneck = maxmin(float('-inf'), float('inf'))


def dual(base):
    class Dual:
        def __init__(self, p, r):
            self.p = p
            self.r = r
        def __eq__(self, y):  return self.p == y.p and self.r == y.r
        def __hash__(self):   return hash((self.p, self.r))
        def __iter__(self):   return iter((self.p, self.r))
        def __repr__(self):   return f'⟨{self.p}, {self.r}⟩'
        def __add__(self, y): return self.__class__(self.p + y.p, self.r + y.r)
        def star(self):
            ps = base.star(self.p)
            return self.__class__(ps, ps * self.r * ps)
        def __mul__(self, y):
            p1,r1 = self.p, self.r
            p2,r2 = y.p, y.r
            return self.__class__(p1*p2, p1*r2 + r1*p2)
        def multiplicity(self, m):
            return self.__class__(base.multiplicity(self.p, m),
                                  base.multiplicity(self.r, m))

    Dual.zero = Dual(base.zero, base.zero)
    Dual.one = Dual(base.one, base.zero)
    return Dual


class String(Wrapped):

    def star(self): return String.one

    def __add__(self, other):
        if other is self.zero: return self
        if self is self.zero: return other
        return String(longest_common_prefix(self.x, other.x))

    def __mul__(self, other):
        if other is self.one:  return self
        if self is self.one:   return other
        if other is self.zero: return self.zero
        if self is self.zero:  return self.zero
        return String(self.x + other.x)

    def __truediv__(self, other):
        prefix = longest_common_prefix(self.x, other.x)
        return String(self.x[len(prefix):])


def longest_common_prefix(xs, ys):
    "computes the longest common prefix"
    position = -1
    for n in range(min(len(xs), len(ys))):
        if xs[n] == ys[n]:
            position = n
        else:
            break
    return xs[:position+1]


String.zero = String(None)   # unique "infinity" string
String.one = String("")


class ThreeValuedLogic(Wrapped):
    def __add__(self, other): return max(self, other)
    def __mul__(self, other): return min(self, other)
ThreeValuedLogic.zero = ThreeValuedLogic(-1)
ThreeValuedLogic.one = ThreeValuedLogic(+1)
ThreeValuedLogic.unk = ThreeValuedLogic(0)
